import json
from typing import (
    Dict,
    List,
    Optional,
)

from messagebus import (
    DataType,
    DefaultToken,
    Message,
    MessageBus,
    TokenCreator,
    TokenPriority,
)
from model.exceptions import ErrorLoadingPosTranslations
from ._I18NRepository import I18NRepository
from ._L10NRepository import L10NRepository
from ._POSControllerRepository import POSControllerRepository

ignore_languages = ["pt", "en_UK", "en_US"]
TK_I18N_GETMSG = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group="7",
    number="2",
)
TK_I18N_GETLANGS = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group="7",
    number="4"
)
TK_I18N_GETTABLEJSON = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group="7",
    number="5",
)


class I18N(object):

    def __init__(
        self,
        message_bus: MessageBus,
        pos_controller_repository: POSControllerRepository,
    ) -> None:

        self.message_bus = message_bus
        self.pos_controller_repository = pos_controller_repository

        self._l10n_params: Dict[int, Dict[str, str]] = {}
        self._i18n_labels: Dict[str, Dict[str, str]] = {}

        self._set_i18n_labels_for_each_language()

    def translate_labels(
        self,
        pos_id: int,
        labels: List[str],
    ) -> Dict[str, Optional[str]]:

        if pos_id not in self._l10n_params:
            self._set_l10n_params(pos_id=pos_id)

        dict_words_translated: Dict[str, Optional[str]] = {}
        language = self._get_language(pos_id=pos_id)
        for label in labels:
            translated_label = self._i18n_labels[language].get(
                label.replace("$", "")
            )
            dict_words_translated[label] = translated_label

        return dict_words_translated

    def build_l10n_repository(
        self,
        pos_id: int,
    ) -> L10NRepository:

        if pos_id not in self._l10n_params:
            self._set_l10n_params(pos_id=pos_id)

        return L10NRepository(l10n_params=self._l10n_params[pos_id])

    def build_i18n_repository(
        self,
        pos_id: int,
    ) -> I18NRepository:

        try:
            language = self._get_language(pos_id=pos_id)
            labels = self._get_i18n_labels(language=language)
            return I18NRepository(i18n_labels=labels)

        except (Exception,):
            raise ErrorLoadingPosTranslations()

    def _set_i18n_labels_for_each_language(
        self,
    ) -> None:

        languages = self._get_i18n_languages()
        for language in languages:
            self._set_i18n_table_labels(language=language)

    def _get_i18n_languages(
        self,
    ) -> List[str]:

        message = Message(
            data_type=DataType.param,
            token=TK_I18N_GETLANGS,
        )
        response = self.message_bus.send_message(
            component="I18N",
            message=message,
        )
        if response.token != DefaultToken.TK_SYS_ACK:
            raise RuntimeError("Error trying to get all languages from I18N")

        all_languages = response.data.decode().split("\0")
        languages = []
        for language in all_languages:
            if language and language not in ignore_languages:
                languages.append(language)

        return languages

    def _get_i18n_labels(
        self,
        language: str,
    ) -> Dict[str, str]:

        message = Message(
            data=language.encode(),
            data_type=DataType.param,
            token=TK_I18N_GETTABLEJSON,
        )
        response = self.message_bus.send_message(
            component="I18N",
            message=message,
        )
        if response.token != DefaultToken.TK_SYS_ACK:
            raise RuntimeError(f"Error trying to get table labels {language} from I18N")

        return json.loads(response.data.decode())

    def _set_i18n_table_labels(
        self,
        language: str,
    ) -> None:

        message = Message(
            data=language.encode(),
            data_type=DataType.param,
            token=TK_I18N_GETTABLEJSON,
        )
        response = self.message_bus.send_message(
            component="I18N",
            message=message,
        )
        if response.token != DefaultToken.TK_SYS_ACK:
            raise RuntimeError(f"Error trying to get table labels {language} from I18N")

        all_labels = json.loads(response.data.decode())
        self._i18n_labels[language] = all_labels

    def _get_language(
        self,
        pos_id: int,
    ) -> str:

        pos_model = self.pos_controller_repository.get_model(pos_id=pos_id)
        if pos_model:
            language = pos_model.find("Language")
            if language is not None:
                return language.get("name", "")

        raise RuntimeError(f"Error trying to get Language, pos id {pos_id} does not exist in PosController cache")

    def _set_l10n_params(
        self,
        pos_id: int,
    ) -> None:

        l10n_params = [
            "L10_PYTHON_DATE_FORMAT",
            "L10_PYTHON_DATETIME_FORMAT",
            "L10N_DECIMAL_PLACES",
            "L10N_CURRENCY_DECIMALS",
            "L10N_CURRENCY_SYMBOL",
            "L10N_DECIMALS_SEPARATOR",
            "L10N_THOUSANDS_SEPARATOR",
            "L10N_CURRENCY_SYMBOL_POSITION",
        ]

        language = self._get_language(pos_id=pos_id)
        l10n_param = language + "\0"
        l10n_param += "\0".join(l10n_params)

        message = Message(
            token=TK_I18N_GETMSG,
            data_type=DataType.param,
            data=l10n_param.encode(),
        )
        response = self.message_bus.send_message(
            component="I18N",
            message=message,
        )
        if response.token != DefaultToken.TK_SYS_ACK:
            raise RuntimeError(f"Invalid token from I18n: TK_I18N_GETMSG - {response.data.decode()}")

        l10n_data = self._remove_language(data=response.data.decode())
        self._l10n_params[pos_id] = {}
        count = 0
        for _ in range(len(l10n_params)):
            self._l10n_params[pos_id][l10n_params[count]] = l10n_data[count]
            count += 1

    @staticmethod
    def _remove_language(
        data: str,
    ) -> List[str]:

        return data.split("\0")[1:]



from sysactions import (
    action,
    get_model,
    get_posot,
    check_current_order,
    show_messagebox,
)
from posot import OrderTakerException

from actions.discount.clean_discounts import clean_order_payments
from actions.order.must_sync_ui import must_sync_ui


@action
def do_back_from_total(pos_id, void_reason=5, *args):
    model = get_model(pos_id)
    order = model.find("CurrentOrder/Order")

    try:
        must_sync_ui(
            pos_id=pos_id,
            different_than="TOTALED",
        )
        clean_order_payments(
            pos_id=pos_id,
            order=order,
            check_manual_discounts=True
        )
        check_current_order(pos_id, model=model)
        get_posot(model).reopenOrder(int(pos_id))

    except OrderTakerException as ex:
        show_messagebox(
            posid=pos_id,
            message="$ERROR_CODE_INFO|%d|%s" % (ex.getErrorCode(), ex.getErrorDescr()),
        )



from __future__ import annotations

from typing import Any
from typing import Iterable
from typing import Mapping
from typing import Optional
from typing import Union

from ..sql._typing import _DDLColumnArgument
from ..sql.elements import DQLDMLClauseElement
from ..sql.schema import ColumnCollectionConstraint
from ..sql.schema import Index


_OnConflictConstraintT = Union[str, ColumnCollectionConstraint, Index, None]
_OnConflictIndexElementsT = Optional[Iterable[_DDLColumnArgument]]
_OnConflictIndexWhereT = Optional[DQLDMLClauseElement]
_OnConflictSetT = Optional[Mapping[Any, Any]]
_OnConflictWhereT = Union[DQLDMLClauseElement, str, None]



from threading import Thread, Condition
from sysactions import set_custom
from actions.pos import pos_must_be_blocked


class UpdatePosBlockedState(Thread):
    def __init__(self, pos_list):
        super(UpdatePosBlockedState, self).__init__()
        self.pos_list = pos_list
        self.daemon = True
        self.thread_condition = Condition()

    def run(self):
        for pos_id in self.pos_list:
            set_custom(pos_id, 'POS_MUST_BE_BLOCKED', False)

        while True:
            for pos_id in self.pos_list:
                if pos_must_be_blocked(pos_id):
                    set_custom(pos_id, 'POS_MUST_BE_BLOCKED', True)

            with self.thread_condition:
                self.thread_condition.wait(5 * 60)



# onrc.py - functions for handling Romanian ONRC numbers
# coding: utf-8
#
# Copyright (C) 2020 Dimitrios Josef Moustos
# Copyright (C) 2020 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""ONRC (Ordine din Registrul Comerţului, Romanian Trade Register identifier).

All businesses in Romania have the to register with the National Trade
Register Office to receive a registration number. The number contains
information about the type of company, county, a sequence number and
registration year. This number can change when registration information
changes.

>>> validate('J52/750/2012')
'J52/750/2012'
>>> validate('X52/750/2012')
Traceback (most recent call last):
    ...
InvalidComponent: ...
"""

import datetime
import re

from stdnum.exceptions import *
from stdnum.util import clean


# These characters should all be replaced by slashes
_cleanup_re = re.compile(r'[ /\\-]+')

# This pattern should match numbers that for some reason have a full date
# as last field
_onrc_fulldate_re = re.compile(r'^([A-Z][0-9]+/[0-9]+/)\d{2}[.]\d{2}[.](\d{4})$')

# This pattern should match all valid numbers
_onrc_re = re.compile(r'^[A-Z][0-9]+/[0-9]+/[0-9]+$')

# List of valid counties
_counties = set(list(range(1, 41)) + [51, 52])


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    number = _cleanup_re.sub('/', clean(number).upper().strip())
    # remove optional slash between first letter and county digits
    if number[1:2] == '/':
        number = number[:1] + number[2:]
    # normalise county number to two digits
    if number[2:3] == '/':
        number = number[:1] + '0' + number[1:]
    # convert trailing full date to year only
    m = _onrc_fulldate_re.match(number)
    if m:
        number = ''.join(m.groups())
    return number


def validate(number):
    """Check if the number is a valid ONRC."""
    number = compact(number)
    if not _onrc_re.match(number):
        raise InvalidFormat()
    if number[:1] not in 'JFC':
        raise InvalidComponent()
    county, serial, year = number[1:].split('/')
    if len(serial) > 5:
        raise InvalidLength()
    if len(county) not in (1, 2) or int(county) not in _counties:
        raise InvalidComponent()
    if len(year) != 4:
        raise InvalidLength()
    if int(year) < 1990 or int(year) > datetime.date.today().year:
        raise InvalidComponent()
    return number


def is_valid(number):
    """Check if the number is a valid ONRC."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



from typing import Dict


class ViewConfiguration(object):
    def __init__(self, name, view_type, log_level, extra_config):
        # type: (str, str, str, Dict[str, str]) -> None
        self.name = name
        self.type = view_type
        self.log_level = log_level
        self.extra_config = extra_config



# bic.py - functions for handling ISO 9362 Business identifier codes
#
# Copyright (C) 2015 Lifealike Ltd
# Copyright (C) 2017-2018 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""BIC (ISO 9362 Business identifier codes).

An ISO 9362 identifier (also: BIC, BEI, or SWIFT code) uniquely
identifies an institution. They are commonly used to route financial
transactions.

The code consists of a 4 letter institution code, a 2 letter country code,
and a 2 character location code, optionally followed by a three character
branch code.

>>> validate('AGRIFRPP882')
'AGRIFRPP882'
>>> validate('ABNA BE 2A')
'ABNABE2A'
>>> validate('AGRIFRPP')
'AGRIFRPP'
>>> validate('AGRIFRPP8')
Traceback (most recent call last):
    ...
InvalidLength: ..
>>> validate('AGRIF2PP')  # country code can't contain digits
Traceback (most recent call last):
    ...
InvalidFormat: ..
>>> format('agriFRPP')  # conventionally caps
'AGRIFRPP'

"""

import re

from stdnum.exceptions import *
from stdnum.util import clean


_bic_re = re.compile(r'^[A-Z]{6}[0-9A-Z]{2}([0-9A-Z]{3})?$')


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any surrounding whitespace."""
    return clean(number, ' -').strip().upper()


def validate(number):
    """Check if the number is a valid routing number. This checks the length
    and characters in each position."""
    number = compact(number)
    if len(number) not in (8, 11):
        raise InvalidLength()
    match = _bic_re.search(number)
    if not match:
        raise InvalidFormat()
    return number


def is_valid(number):
    """Check if the number provided is a valid BIC."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    return compact(number)



import json
import logging

from bustoken import TK_LOYALTY_GET_TOTALED_MESSAGE
from msgbus import (
    FM_PARAM,
    TK_SYS_ACK,
)
from sysactions import send_message
from typing import Optional
from actions.models import (
    LoyaltyTranslatedMessage,
    LoyaltyTotaledMessage,
)
from actions.util import get_pos_language

logger = logging.getLogger("PosActions")


def get_loyalty_totaled_message(
    pos_id,         # type: str
    customer_id,    # type: str
):
    # type: (...) -> Optional[LoyaltyTotaledMessage]

    data = {
        "posId": pos_id,
        "customerId": customer_id,
    }

    try:
        logger.info("Start getting loyalty totaled message")
        response = send_message(
            dest_name="Loyalty",
            token=TK_LOYALTY_GET_TOTALED_MESSAGE,
            format=FM_PARAM,
            data=json.dumps(data),
        )
        if response.token == TK_SYS_ACK:
            response_data = json.loads(response.data)
            loyalty_translated_message = LoyaltyTranslatedMessage(
                pt=response_data.get("totalizationMessage").get("pt"),
                en=response_data.get("totalizationMessage").get("en"),
                es=response_data.get("totalizationMessage").get("es"),
            )
            language = get_pos_language(pos_id=pos_id)[0:2]
            loyalty_totaled_message = LoyaltyTotaledMessage(loyalty_translated_message=loyalty_translated_message)
            translated_message = getattr(loyalty_totaled_message.totalization_message, language)
            if translated_message:
                return translated_message
        logger.info("TK_LOYALTY_GET_TOTALED_MESSAGE returned NAK")
        return None
    except Exception as _:
        logger.exception("Error getting loyalty totaled message")
        return None



import persistence
import datetime
import hmac

from persistence import Driver

_MWPOS_BASE_HMAC = 0x18122008


class UserHelper:
    def __init__(self, mbcontext):
        # type: (msgbus.MBEasyContext) -> None
        self.mbcontext = mbcontext
        self.conn = None  # type: persistence.Connection

    def __enter__(self):
        self.conn = Driver().open(self.mbcontext)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.conn:
            self.conn.close()
        return exc_val

    def get_all_users(self):
        # type: () -> dict
        return {x.get_entry(0): (x.get_entry(1), x.get_entry(2), x.get_entry(3), x.get_entry(4))
                for x in self.conn.select("""select UserId, LongName, Level, Status from Users""")}

    def get_users_by_level(self, level):
        # type: () -> dict
        return {x.get_entry(0): (x.get_entry(1), x.get_entry(2)) for x in self.conn.select("""select UserId, UserName, case when FingerPrintMinutiae is NULL then 0 else 1 end as HasPrinterPrint from Users where level='%s' and status=0""" % (level))}

    def get_inactive_users_by_level(self, level):
        # type: () -> dict
        return {x.get_entry(0): (x.get_entry(1), x.get_entry(2)) for x in self.conn.select("""select UserId, UserName, case when FingerPrintMinutiae is NULL then 0 else 1 end as HasPrinterPrint from Users where level='%s' and status=1""" % (level))}

    def get_user_by_id(self, userid):
        return {x.get_entry(0): x.get_entry(1) for x in self.conn.select("select UserId, UserName from Users where UserId = '%s'" % userid)}

    def add_user(self, userid, username, passwd, admin, fmd_base64):
        # type: (str, etree.XML.ElementTree.Element, str) -> int
        calc = hmac.new(str(int(userid) + _MWPOS_BASE_HMAC))
        calc.update(passwd)
        the_passwd = calc.hexdigest()
        if not fmd_base64:
            return {self.conn.query(
                "insert into Users(UserId, UserName, LongName, Password, Level, Status, AdmissionDate, PayRate) values ('%s','%s','%s','%s',%d,%d,'%s',%f)" % (
                    userid, username, username, the_passwd, 20 if admin else 0, 0, datetime.date.today(), 8.5))}
        else:
            return {self.conn.query(
                "insert into Users(UserId, UserName, LongName, Password, Level, Status, AdmissionDate, PayRate, FingerPrintMinutiae) values ('%s','%s','%s','%s',%d,%d,'%s',%f,x'%s')" % (
                    userid, username, username, the_passwd, 20 if admin else 0, 0, datetime.date.today(), 8.5, fmd_base64.encode("hex")))}

    def change_pass(self, userid, passwd):
        # type: (str, str) -> int
        calc = hmac.new(str(int(userid) + _MWPOS_BASE_HMAC))
        calc.update(passwd)
        the_passwd = calc.hexdigest()
        return self.conn.query("update Users set Password = '%s' where UserId = '%s'" % (the_passwd, userid))

    def remove_user(self, userid):
        return {self.conn.query("delete from Users where UserId = %s" % userid)}

    def activate_user(self, userid, op):
        # op = 0 Activate - 1 Deactivate
        return {self.conn.query(
            "update Users set Status = '%d' where UserId = '%s'" % (op, userid))}

    def change_level_user(self, userid, level):
        return {self.conn.query("update Users set level = '%s' where UserId = '%s'" % (level, userid))}

    def get_level_user(self, userid):
        return {'level': x.get_entry(0) for x in self.conn.select("select level from Users where UserId = '%s'" % (userid))}

    def get_user_session(self, posid):
        return {x.get_entry(0): x.get_entry(1) for x in self.conn.select("""
          select us.sessionid, u.level from posctrl.UserSession us join Users u on u.userid = us.operatorid where us.closetime IS NULL and us.posid = '%s'""" % (posid))}

    def get_all_logged_users(self):
        query = """select u.userid 
                     from posctrl.UserSession us 
                     join Users u on u.userid = us.operatorid 
                    where us.closetime IS NULL"""

        select = self.conn.select(query)
        return {x.get_entry(0) for x in select}



# -*- coding: utf-8 -*-

from json import JSONEncoder

from ._PosVersions import PosVersions
from ._UpdateControl import UpdateControl


class ModelJsonEncoder(JSONEncoder):
    def default(self, o):
        if isinstance(o, PosVersions):
            return (
                {
                    "versions": {
                        "catalog": o.catalog,
                        "user": o.user,
                        "src": o.src,
                        "core": o.core,
                        "loader": o.loader,
                        "binaries": o.binaries
                    },
                    "dates": {
                        "catalog": o.catalog_date,
                        "user": o.user_date,
                        "src": o.src_date,
                        "core": o.core_date,
                        "loader": o.loader_date,
                        "binaries": o.binaries_date
                    },
                    "updateControl": o.update_control,
                    "menuVersion": o.menu_version,
                    "menuName": o.menu_name,
                }
            )

        if isinstance(o, UpdateControl):
            return (
                {
                    "binaries": o.binaries,
                    "catalog": o.catalog,
                    "loader": o.loader,
                    "user": o.user,
                    "updating": o.updating
                }
            )

        return super(ModelJsonEncoder, self).default(o)



"""A PEP 517 interface to setuptools

Previously, when a user or a command line tool (let's call it a "frontend")
needed to make a request of setuptools to take a certain action, for
example, generating a list of installation requirements, the frontend
would call "setup.py egg_info" or "setup.py bdist_wheel" on the command line.

PEP 517 defines a different method of interfacing with setuptools. Rather
than calling "setup.py" directly, the frontend should:

  1. Set the current directory to the directory with a setup.py file
  2. Import this module into a safe python interpreter (one in which
     setuptools can potentially set global variables or crash hard).
  3. Call one of the functions defined in PEP 517.

What each function does is defined in PEP 517. However, here is a "casual"
definition of the functions (this definition should not be relied on for
bug reports or API stability):

  - `build_wheel`: build a wheel in the folder and return the basename
  - `get_requires_for_build_wheel`: get the `setup_requires` to build
  - `prepare_metadata_for_build_wheel`: get the `install_requires`
  - `build_sdist`: build an sdist in the folder and return the basename
  - `get_requires_for_build_sdist`: get the `setup_requires` to build

Again, this is not a formal definition! Just a "taste" of the module.
"""

import io
import os
import shlex
import sys
import tokenize
import shutil
import contextlib
import tempfile
import warnings
from pathlib import Path
from typing import Dict, Iterator, List, Optional, Union

import setuptools
import distutils
from . import errors
from ._path import same_path
from ._reqs import parse_strings
from .warnings import SetuptoolsDeprecationWarning
from distutils.util import strtobool


__all__ = [
    'get_requires_for_build_sdist',
    'get_requires_for_build_wheel',
    'prepare_metadata_for_build_wheel',
    'build_wheel',
    'build_sdist',
    'get_requires_for_build_editable',
    'prepare_metadata_for_build_editable',
    'build_editable',
    '__legacy__',
    'SetupRequirementsError',
]

SETUPTOOLS_ENABLE_FEATURES = os.getenv("SETUPTOOLS_ENABLE_FEATURES", "").lower()
LEGACY_EDITABLE = "legacy-editable" in SETUPTOOLS_ENABLE_FEATURES.replace("_", "-")


class SetupRequirementsError(BaseException):
    def __init__(self, specifiers):
        self.specifiers = specifiers


class Distribution(setuptools.dist.Distribution):
    def fetch_build_eggs(self, specifiers):
        specifier_list = list(parse_strings(specifiers))

        raise SetupRequirementsError(specifier_list)

    @classmethod
    @contextlib.contextmanager
    def patch(cls):
        """
        Replace
        distutils.dist.Distribution with this class
        for the duration of this context.
        """
        orig = distutils.core.Distribution
        distutils.core.Distribution = cls
        try:
            yield
        finally:
            distutils.core.Distribution = orig


@contextlib.contextmanager
def no_install_setup_requires():
    """Temporarily disable installing setup_requires

    Under PEP 517, the backend reports build dependencies to the frontend,
    and the frontend is responsible for ensuring they're installed.
    So setuptools (acting as a backend) should not try to install them.
    """
    orig = setuptools._install_setup_requires
    setuptools._install_setup_requires = lambda attrs: None
    try:
        yield
    finally:
        setuptools._install_setup_requires = orig


def _get_immediate_subdirectories(a_dir):
    return [
        name for name in os.listdir(a_dir) if os.path.isdir(os.path.join(a_dir, name))
    ]


def _file_with_extension(directory, extension):
    matching = (f for f in os.listdir(directory) if f.endswith(extension))
    try:
        (file,) = matching
    except ValueError:
        raise ValueError(
            'No distribution was found. Ensure that `setup.py` '
            'is not empty and that it calls `setup()`.'
        ) from None
    return file


def _open_setup_script(setup_script):
    if not os.path.exists(setup_script):
        # Supply a default setup.py
        return io.StringIO("from setuptools import setup; setup()")

    return tokenize.open(setup_script)


@contextlib.contextmanager
def suppress_known_deprecation():
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', 'setup.py install is deprecated')
        yield


_ConfigSettings = Optional[Dict[str, Union[str, List[str], None]]]
"""
Currently the user can run::

    pip install -e . --config-settings key=value
    python -m build -C--key=value -C key=value

- pip will pass both key and value as strings and overwriting repeated keys
  (pypa/pip#11059).
- build will accumulate values associated with repeated keys in a list.
  It will also accept keys with no associated value.
  This means that an option passed by build can be ``str | list[str] | None``.
- PEP 517 specifies that ``config_settings`` is an optional dict.
"""


class _ConfigSettingsTranslator:
    """Translate ``config_settings`` into distutils-style command arguments.
    Only a limited number of options is currently supported.
    """

    # See pypa/setuptools#1928 pypa/setuptools#2491

    def _get_config(self, key: str, config_settings: _ConfigSettings) -> List[str]:
        """
        Get the value of a specific key in ``config_settings`` as a list of strings.

        >>> fn = _ConfigSettingsTranslator()._get_config
        >>> fn("--global-option", None)
        []
        >>> fn("--global-option", {})
        []
        >>> fn("--global-option", {'--global-option': 'foo'})
        ['foo']
        >>> fn("--global-option", {'--global-option': ['foo']})
        ['foo']
        >>> fn("--global-option", {'--global-option': 'foo'})
        ['foo']
        >>> fn("--global-option", {'--global-option': 'foo bar'})
        ['foo', 'bar']
        """
        cfg = config_settings or {}
        opts = cfg.get(key) or []
        return shlex.split(opts) if isinstance(opts, str) else opts

    def _global_args(self, config_settings: _ConfigSettings) -> Iterator[str]:
        """
        Let the user specify ``verbose`` or ``quiet`` + escape hatch via
        ``--global-option``.
        Note: ``-v``, ``-vv``, ``-vvv`` have similar effects in setuptools,
        so we just have to cover the basic scenario ``-v``.

        >>> fn = _ConfigSettingsTranslator()._global_args
        >>> list(fn(None))
        []
        >>> list(fn({"verbose": "False"}))
        ['-q']
        >>> list(fn({"verbose": "1"}))
        ['-v']
        >>> list(fn({"--verbose": None}))
        ['-v']
        >>> list(fn({"verbose": "true", "--global-option": "-q --no-user-cfg"}))
        ['-v', '-q', '--no-user-cfg']
        >>> list(fn({"--quiet": None}))
        ['-q']
        """
        cfg = config_settings or {}
        falsey = {"false", "no", "0", "off"}
        if "verbose" in cfg or "--verbose" in cfg:
            level = str(cfg.get("verbose") or cfg.get("--verbose") or "1")
            yield ("-q" if level.lower() in falsey else "-v")
        if "quiet" in cfg or "--quiet" in cfg:
            level = str(cfg.get("quiet") or cfg.get("--quiet") or "1")
            yield ("-v" if level.lower() in falsey else "-q")

        yield from self._get_config("--global-option", config_settings)

    def __dist_info_args(self, config_settings: _ConfigSettings) -> Iterator[str]:
        """
        The ``dist_info`` command accepts ``tag-date`` and ``tag-build``.

        .. warning::
           We cannot use this yet as it requires the ``sdist`` and ``bdist_wheel``
           commands run in ``build_sdist`` and ``build_wheel`` to reuse the egg-info
           directory created in ``prepare_metadata_for_build_wheel``.

        >>> fn = _ConfigSettingsTranslator()._ConfigSettingsTranslator__dist_info_args
        >>> list(fn(None))
        []
        >>> list(fn({"tag-date": "False"}))
        ['--no-date']
        >>> list(fn({"tag-date": None}))
        ['--no-date']
        >>> list(fn({"tag-date": "true", "tag-build": ".a"}))
        ['--tag-date', '--tag-build', '.a']
        """
        cfg = config_settings or {}
        if "tag-date" in cfg:
            val = strtobool(str(cfg["tag-date"] or "false"))
            yield ("--tag-date" if val else "--no-date")
        if "tag-build" in cfg:
            yield from ["--tag-build", str(cfg["tag-build"])]

    def _editable_args(self, config_settings: _ConfigSettings) -> Iterator[str]:
        """
        The ``editable_wheel`` command accepts ``editable-mode=strict``.

        >>> fn = _ConfigSettingsTranslator()._editable_args
        >>> list(fn(None))
        []
        >>> list(fn({"editable-mode": "strict"}))
        ['--mode', 'strict']
        """
        cfg = config_settings or {}
        mode = cfg.get("editable-mode") or cfg.get("editable_mode")
        if not mode:
            return
        yield from ["--mode", str(mode)]

    def _arbitrary_args(self, config_settings: _ConfigSettings) -> Iterator[str]:
        """
        Users may expect to pass arbitrary lists of arguments to a command
        via "--global-option" (example provided in PEP 517 of a "escape hatch").

        >>> fn = _ConfigSettingsTranslator()._arbitrary_args
        >>> list(fn(None))
        []
        >>> list(fn({}))
        []
        >>> list(fn({'--build-option': 'foo'}))
        ['foo']
        >>> list(fn({'--build-option': ['foo']}))
        ['foo']
        >>> list(fn({'--build-option': 'foo'}))
        ['foo']
        >>> list(fn({'--build-option': 'foo bar'}))
        ['foo', 'bar']
        >>> list(fn({'--global-option': 'foo'}))
        []
        """
        yield from self._get_config("--build-option", config_settings)


class _BuildMetaBackend(_ConfigSettingsTranslator):
    def _get_build_requires(self, config_settings, requirements):
        sys.argv = [
            *sys.argv[:1],
            *self._global_args(config_settings),
            "egg_info",
        ]
        try:
            with Distribution.patch():
                self.run_setup()
        except SetupRequirementsError as e:
            requirements += e.specifiers

        return requirements

    def run_setup(self, setup_script='setup.py'):
        # Note that we can reuse our build directory between calls
        # Correctness comes first, then optimization later
        __file__ = os.path.abspath(setup_script)
        __name__ = '__main__'

        with _open_setup_script(__file__) as f:
            code = f.read().replace(r'\r\n', r'\n')

        try:
            exec(code, locals())
        except SystemExit as e:
            if e.code:
                raise
            # We ignore exit code indicating success
            SetuptoolsDeprecationWarning.emit(
                "Running `setup.py` directly as CLI tool is deprecated.",
                "Please avoid using `sys.exit(0)` or similar statements "
                "that don't fit in the paradigm of a configuration file.",
                see_url="https://blog.ganssle.io/articles/2021/10/"
                "setup-py-deprecated.html",
            )

    def get_requires_for_build_wheel(self, config_settings=None):
        return self._get_build_requires(config_settings, requirements=['wheel'])

    def get_requires_for_build_sdist(self, config_settings=None):
        return self._get_build_requires(config_settings, requirements=[])

    def _bubble_up_info_directory(self, metadata_directory: str, suffix: str) -> str:
        """
        PEP 517 requires that the .dist-info directory be placed in the
        metadata_directory. To comply, we MUST copy the directory to the root.

        Returns the basename of the info directory, e.g. `proj-0.0.0.dist-info`.
        """
        info_dir = self._find_info_directory(metadata_directory, suffix)
        if not same_path(info_dir.parent, metadata_directory):
            shutil.move(str(info_dir), metadata_directory)
            # PEP 517 allow other files and dirs to exist in metadata_directory
        return info_dir.name

    def _find_info_directory(self, metadata_directory: str, suffix: str) -> Path:
        for parent, dirs, _ in os.walk(metadata_directory):
            candidates = [f for f in dirs if f.endswith(suffix)]

            if len(candidates) != 0 or len(dirs) != 1:
                assert len(candidates) == 1, f"Multiple {suffix} directories found"
                return Path(parent, candidates[0])

        msg = f"No {suffix} directory found in {metadata_directory}"
        raise errors.InternalError(msg)

    def prepare_metadata_for_build_wheel(
        self, metadata_directory, config_settings=None
    ):
        sys.argv = [
            *sys.argv[:1],
            *self._global_args(config_settings),
            "dist_info",
            "--output-dir",
            metadata_directory,
            "--keep-egg-info",
        ]
        with no_install_setup_requires():
            self.run_setup()

        self._bubble_up_info_directory(metadata_directory, ".egg-info")
        return self._bubble_up_info_directory(metadata_directory, ".dist-info")

    def _build_with_temp_dir(
        self,
        setup_command,
        result_extension,
        result_directory,
        config_settings,
        arbitrary_args=(),
    ):
        result_directory = os.path.abspath(result_directory)

        # Build in a temporary directory, then copy to the target.
        os.makedirs(result_directory, exist_ok=True)
        temp_opts = {"prefix": ".tmp-", "dir": result_directory}

        with tempfile.TemporaryDirectory(**temp_opts) as tmp_dist_dir:
            sys.argv = [
                *sys.argv[:1],
                *self._global_args(config_settings),
                *setup_command,
                "--dist-dir",
                tmp_dist_dir,
                *arbitrary_args,
            ]
            with no_install_setup_requires():
                self.run_setup()

            result_basename = _file_with_extension(tmp_dist_dir, result_extension)
            result_path = os.path.join(result_directory, result_basename)
            if os.path.exists(result_path):
                # os.rename will fail overwriting on non-Unix.
                os.remove(result_path)
            os.rename(os.path.join(tmp_dist_dir, result_basename), result_path)

        return result_basename

    def build_wheel(
        self, wheel_directory, config_settings=None, metadata_directory=None
    ):
        with suppress_known_deprecation():
            return self._build_with_temp_dir(
                ['bdist_wheel'],
                '.whl',
                wheel_directory,
                config_settings,
                self._arbitrary_args(config_settings),
            )

    def build_sdist(self, sdist_directory, config_settings=None):
        return self._build_with_temp_dir(
            ['sdist', '--formats', 'gztar'], '.tar.gz', sdist_directory, config_settings
        )

    def _get_dist_info_dir(self, metadata_directory: Optional[str]) -> Optional[str]:
        if not metadata_directory:
            return None
        dist_info_candidates = list(Path(metadata_directory).glob("*.dist-info"))
        assert len(dist_info_candidates) <= 1
        return str(dist_info_candidates[0]) if dist_info_candidates else None

    if not LEGACY_EDITABLE:
        # PEP660 hooks:
        # build_editable
        # get_requires_for_build_editable
        # prepare_metadata_for_build_editable
        def build_editable(
            self, wheel_directory, config_settings=None, metadata_directory=None
        ):
            # XXX can or should we hide our editable_wheel command normally?
            info_dir = self._get_dist_info_dir(metadata_directory)
            opts = ["--dist-info-dir", info_dir] if info_dir else []
            cmd = ["editable_wheel", *opts, *self._editable_args(config_settings)]
            with suppress_known_deprecation():
                return self._build_with_temp_dir(
                    cmd, ".whl", wheel_directory, config_settings
                )

        def get_requires_for_build_editable(self, config_settings=None):
            return self.get_requires_for_build_wheel(config_settings)

        def prepare_metadata_for_build_editable(
            self, metadata_directory, config_settings=None
        ):
            return self.prepare_metadata_for_build_wheel(
                metadata_directory, config_settings
            )


class _BuildMetaLegacyBackend(_BuildMetaBackend):
    """Compatibility backend for setuptools

    This is a version of setuptools.build_meta that endeavors
    to maintain backwards
    compatibility with pre-PEP 517 modes of invocation. It
    exists as a temporary
    bridge between the old packaging mechanism and the new
    packaging mechanism,
    and will eventually be removed.
    """

    def run_setup(self, setup_script='setup.py'):
        # In order to maintain compatibility with scripts assuming that
        # the setup.py script is in a directory on the PYTHONPATH, inject
        # '' into sys.path. (pypa/setuptools#1642)
        sys_path = list(sys.path)  # Save the original path

        script_dir = os.path.dirname(os.path.abspath(setup_script))
        if script_dir not in sys.path:
            sys.path.insert(0, script_dir)

        # Some setup.py scripts (e.g. in pygame and numpy) use sys.argv[0] to
        # get the directory of the source code. They expect it to refer to the
        # setup.py script.
        sys_argv_0 = sys.argv[0]
        sys.argv[0] = setup_script

        try:
            super().run_setup(setup_script=setup_script)
        finally:
            # While PEP 517 frontends should be calling each hook in a fresh
            # subprocess according to the standard (and thus it should not be
            # strictly necessary to restore the old sys.path), we'll restore
            # the original path so that the path manipulation does not persist
            # within the hook after run_setup is called.
            sys.path[:] = sys_path
            sys.argv[0] = sys_argv_0


# The primary backend
_BACKEND = _BuildMetaBackend()

get_requires_for_build_wheel = _BACKEND.get_requires_for_build_wheel
get_requires_for_build_sdist = _BACKEND.get_requires_for_build_sdist
prepare_metadata_for_build_wheel = _BACKEND.prepare_metadata_for_build_wheel
build_wheel = _BACKEND.build_wheel
build_sdist = _BACKEND.build_sdist

if not LEGACY_EDITABLE:
    get_requires_for_build_editable = _BACKEND.get_requires_for_build_editable
    prepare_metadata_for_build_editable = _BACKEND.prepare_metadata_for_build_editable
    build_editable = _BACKEND.build_editable


# The legacy backend
__legacy__ = _BuildMetaLegacyBackend()



# -*- coding: utf-8 -*-
from logging import Logger

import cfgtools
from messagebus import MessageBus
from messagehandler import MessageHandlerBuilder
from msgbus import MBEasyContext
from typing import List

from _ProxyMessageHandler import ProxyMessageHandler
from model import Kds


class NKDSCtrlMessageHandlerBuilder(MessageHandlerBuilder):
    def __init__(
        self,
        comp_name,    # type: str
        mb_context,   # type: MBEasyContext
        message_bus,  # type: MessageBus
        config,       # type: cfgtools.Group
        logger        # type: Logger
    ):
        # type: (...) -> None  # noqa

        self.comp_name = comp_name
        self.mb_context = mb_context
        self.message_bus = message_bus
        self.config = config
        self.logger = logger

        self.proxy = None

    def build_singletons(self):
        self.proxy = ProxyMessageHandler(
            self.comp_name,
            self.mb_context,
            self.message_bus,
            self.config,
            self.logger
        )

    def build_message_handler(self, message, event):
        return self.proxy

    def destroy_message_handler(self, message_handler):
        return

    def destroy_singletons(self):
        return



from ._HTMLTextFormatter import HTMLTextFormatter
from ._LineParser import (
    LineParser,
    JSON,
)


COMMANDS = {
    "{CMD_BREAK_LINE}": " ",
    "{CMD_SEPARATOR_EQUALS}": "=",
    "{CMD_SEPARATOR_DASH}": "-",
    "{CMD_NEW_PAGE}": "-",
}


class CommandLineParser(LineParser):

    def __init__(
        self,       # type: CommandLineParser
        columns,    # type: int
    ):
        # type: (...) -> None

        super(CommandLineParser, self).__init__(columns)

    def parse(
        self,    # type: CommandLineParser
        line,    # type: JSON
    ):

        parsed_command = self._parse_command(command=line.get("command"))
        html_text_formatter = HTMLTextFormatter(
            customization=line.get("customization"),
            columns=self.columns,
        )

        return [html_text_formatter.format(text=parsed_command)]

    def _parse_command(
        self,       # type: CommandLineParser
        command,    # type: str
    ):
        # type: (...) -> str

        parsed_command = COMMANDS.get(command, " ")
        return parsed_command * self.columns



data = (
'ggwem',    # 0x00
'ggweb',    # 0x01
'ggwebs',    # 0x02
'ggwes',    # 0x03
'ggwess',    # 0x04
'ggweng',    # 0x05
'ggwej',    # 0x06
'ggwec',    # 0x07
'ggwek',    # 0x08
'ggwet',    # 0x09
'ggwep',    # 0x0a
'ggweh',    # 0x0b
'ggwi',    # 0x0c
'ggwig',    # 0x0d
'ggwigg',    # 0x0e
'ggwigs',    # 0x0f
'ggwin',    # 0x10
'ggwinj',    # 0x11
'ggwinh',    # 0x12
'ggwid',    # 0x13
'ggwil',    # 0x14
'ggwilg',    # 0x15
'ggwilm',    # 0x16
'ggwilb',    # 0x17
'ggwils',    # 0x18
'ggwilt',    # 0x19
'ggwilp',    # 0x1a
'ggwilh',    # 0x1b
'ggwim',    # 0x1c
'ggwib',    # 0x1d
'ggwibs',    # 0x1e
'ggwis',    # 0x1f
'ggwiss',    # 0x20
'ggwing',    # 0x21
'ggwij',    # 0x22
'ggwic',    # 0x23
'ggwik',    # 0x24
'ggwit',    # 0x25
'ggwip',    # 0x26
'ggwih',    # 0x27
'ggyu',    # 0x28
'ggyug',    # 0x29
'ggyugg',    # 0x2a
'ggyugs',    # 0x2b
'ggyun',    # 0x2c
'ggyunj',    # 0x2d
'ggyunh',    # 0x2e
'ggyud',    # 0x2f
'ggyul',    # 0x30
'ggyulg',    # 0x31
'ggyulm',    # 0x32
'ggyulb',    # 0x33
'ggyuls',    # 0x34
'ggyult',    # 0x35
'ggyulp',    # 0x36
'ggyulh',    # 0x37
'ggyum',    # 0x38
'ggyub',    # 0x39
'ggyubs',    # 0x3a
'ggyus',    # 0x3b
'ggyuss',    # 0x3c
'ggyung',    # 0x3d
'ggyuj',    # 0x3e
'ggyuc',    # 0x3f
'ggyuk',    # 0x40
'ggyut',    # 0x41
'ggyup',    # 0x42
'ggyuh',    # 0x43
'ggeu',    # 0x44
'ggeug',    # 0x45
'ggeugg',    # 0x46
'ggeugs',    # 0x47
'ggeun',    # 0x48
'ggeunj',    # 0x49
'ggeunh',    # 0x4a
'ggeud',    # 0x4b
'ggeul',    # 0x4c
'ggeulg',    # 0x4d
'ggeulm',    # 0x4e
'ggeulb',    # 0x4f
'ggeuls',    # 0x50
'ggeult',    # 0x51
'ggeulp',    # 0x52
'ggeulh',    # 0x53
'ggeum',    # 0x54
'ggeub',    # 0x55
'ggeubs',    # 0x56
'ggeus',    # 0x57
'ggeuss',    # 0x58
'ggeung',    # 0x59
'ggeuj',    # 0x5a
'ggeuc',    # 0x5b
'ggeuk',    # 0x5c
'ggeut',    # 0x5d
'ggeup',    # 0x5e
'ggeuh',    # 0x5f
'ggyi',    # 0x60
'ggyig',    # 0x61
'ggyigg',    # 0x62
'ggyigs',    # 0x63
'ggyin',    # 0x64
'ggyinj',    # 0x65
'ggyinh',    # 0x66
'ggyid',    # 0x67
'ggyil',    # 0x68
'ggyilg',    # 0x69
'ggyilm',    # 0x6a
'ggyilb',    # 0x6b
'ggyils',    # 0x6c
'ggyilt',    # 0x6d
'ggyilp',    # 0x6e
'ggyilh',    # 0x6f
'ggyim',    # 0x70
'ggyib',    # 0x71
'ggyibs',    # 0x72
'ggyis',    # 0x73
'ggyiss',    # 0x74
'ggying',    # 0x75
'ggyij',    # 0x76
'ggyic',    # 0x77
'ggyik',    # 0x78
'ggyit',    # 0x79
'ggyip',    # 0x7a
'ggyih',    # 0x7b
'ggi',    # 0x7c
'ggig',    # 0x7d
'ggigg',    # 0x7e
'ggigs',    # 0x7f
'ggin',    # 0x80
'gginj',    # 0x81
'gginh',    # 0x82
'ggid',    # 0x83
'ggil',    # 0x84
'ggilg',    # 0x85
'ggilm',    # 0x86
'ggilb',    # 0x87
'ggils',    # 0x88
'ggilt',    # 0x89
'ggilp',    # 0x8a
'ggilh',    # 0x8b
'ggim',    # 0x8c
'ggib',    # 0x8d
'ggibs',    # 0x8e
'ggis',    # 0x8f
'ggiss',    # 0x90
'gging',    # 0x91
'ggij',    # 0x92
'ggic',    # 0x93
'ggik',    # 0x94
'ggit',    # 0x95
'ggip',    # 0x96
'ggih',    # 0x97
'na',    # 0x98
'nag',    # 0x99
'nagg',    # 0x9a
'nags',    # 0x9b
'nan',    # 0x9c
'nanj',    # 0x9d
'nanh',    # 0x9e
'nad',    # 0x9f
'nal',    # 0xa0
'nalg',    # 0xa1
'nalm',    # 0xa2
'nalb',    # 0xa3
'nals',    # 0xa4
'nalt',    # 0xa5
'nalp',    # 0xa6
'nalh',    # 0xa7
'nam',    # 0xa8
'nab',    # 0xa9
'nabs',    # 0xaa
'nas',    # 0xab
'nass',    # 0xac
'nang',    # 0xad
'naj',    # 0xae
'nac',    # 0xaf
'nak',    # 0xb0
'nat',    # 0xb1
'nap',    # 0xb2
'nah',    # 0xb3
'nae',    # 0xb4
'naeg',    # 0xb5
'naegg',    # 0xb6
'naegs',    # 0xb7
'naen',    # 0xb8
'naenj',    # 0xb9
'naenh',    # 0xba
'naed',    # 0xbb
'nael',    # 0xbc
'naelg',    # 0xbd
'naelm',    # 0xbe
'naelb',    # 0xbf
'naels',    # 0xc0
'naelt',    # 0xc1
'naelp',    # 0xc2
'naelh',    # 0xc3
'naem',    # 0xc4
'naeb',    # 0xc5
'naebs',    # 0xc6
'naes',    # 0xc7
'naess',    # 0xc8
'naeng',    # 0xc9
'naej',    # 0xca
'naec',    # 0xcb
'naek',    # 0xcc
'naet',    # 0xcd
'naep',    # 0xce
'naeh',    # 0xcf
'nya',    # 0xd0
'nyag',    # 0xd1
'nyagg',    # 0xd2
'nyags',    # 0xd3
'nyan',    # 0xd4
'nyanj',    # 0xd5
'nyanh',    # 0xd6
'nyad',    # 0xd7
'nyal',    # 0xd8
'nyalg',    # 0xd9
'nyalm',    # 0xda
'nyalb',    # 0xdb
'nyals',    # 0xdc
'nyalt',    # 0xdd
'nyalp',    # 0xde
'nyalh',    # 0xdf
'nyam',    # 0xe0
'nyab',    # 0xe1
'nyabs',    # 0xe2
'nyas',    # 0xe3
'nyass',    # 0xe4
'nyang',    # 0xe5
'nyaj',    # 0xe6
'nyac',    # 0xe7
'nyak',    # 0xe8
'nyat',    # 0xe9
'nyap',    # 0xea
'nyah',    # 0xeb
'nyae',    # 0xec
'nyaeg',    # 0xed
'nyaegg',    # 0xee
'nyaegs',    # 0xef
'nyaen',    # 0xf0
'nyaenj',    # 0xf1
'nyaenh',    # 0xf2
'nyaed',    # 0xf3
'nyael',    # 0xf4
'nyaelg',    # 0xf5
'nyaelm',    # 0xf6
'nyaelb',    # 0xf7
'nyaels',    # 0xf8
'nyaelt',    # 0xf9
'nyaelp',    # 0xfa
'nyaelh',    # 0xfb
'nyaem',    # 0xfc
'nyaeb',    # 0xfd
'nyaebs',    # 0xfe
'nyaes',    # 0xff
)



from datetime import datetime
from logging import Logger
from threading import (
    Thread,
    Condition,
)

from mbcontextmessagehandler import MbContextMessageBus

from externals.repository import AlertsRepository
from services.model import Configs


class DatabaseManager(object):
    def __init__(
        self,               # type: DatabaseManager
        message_bus,        # type: MbContextMessageBus
        alerts_repository,  # type: AlertsRepository
        configs,            # type: Configs
        logger,             # type: Logger
    ):
        # type: (...) -> None

        self.message_bus = message_bus
        self.alerts_repository = alerts_repository
        self.start_time = configs.alerts.maintenance_window_start
        self.end_time = configs.alerts.maintenance_window_end
        self.days_to_keep_data = configs.alerts.days_to_keep_data
        self.logger = logger

        self.thread = None
        self.running = False
        self.sleep_condition = Condition()
        self.interval_in_minutes = 30
        self.has_already_applied_cleaning = False

    def start(
        self,  # type: DatabaseManager
    ):
        # type: (...) -> None

        if not self.running:
            self.running = True
            self.thread = Thread(
                target=self._clean_runner,
                name="CleanDatabases",
            )
            self.thread.daemon = True
            self.thread.start()

    def stop(
        self,   # type: DatabaseManager
    ):
        # type: (...) -> None

        if self.running:
            with self.sleep_condition:
                self.running = False
                self.sleep_condition.notify()

    def _clean_runner(
        self,   # type: DatabaseManager
    ):
        # type: (...) -> None

        self._handle_clean_databases()

        clean_only_on_component_start = self.start_time is None or self.end_time is None
        if clean_only_on_component_start:
            self.stop()
            return

        while self.running:
            if self._allowed_to_clean_databases():
                self._handle_clean_databases()

            with self.sleep_condition:
                self.sleep_condition.wait(self.interval_in_minutes * 60)

    def _handle_clean_databases(
        self,   # type: DatabaseManager
    ):
        # type: (...) -> None

        self.logger.info("Start cleaning databases")

        try:
            self.alerts_repository.delete_outdated_events(range_days=self.days_to_keep_data)
            self.alerts_repository.vacuum_database()
            self.has_already_applied_cleaning = True
        except Exception as _:
            self.logger.exception(
                "Error cleaning databases",
                exc_info=True,
            )

        self.logger.info("Databases cleaned with success")

    def _allowed_to_clean_databases(
        self,   # type: DatabaseManager
    ):
        # type: (...) -> bool

        now = datetime.now().time()
        is_inside_clean_window = self.start_time <= now <= self.end_time

        if self.has_already_applied_cleaning and not is_inside_clean_window:
            self.has_already_applied_cleaning = False

        if not self.has_already_applied_cleaning and is_inside_clean_window:
            return True

        return False



from model import (
    EventPublishDetail,
    NewEventTopic,
    AsyncTarget,
)
from model.repository import EventRepository
from opentelemetry import trace
from opentelemetry.trace import Span, SpanKind
from timeutil import Clock
from .dto import (
    PublishEventRequest,
    CannotProcessSyncEvent,
)

tracer = trace.get_tracer(__name__)


class PublishEventInteractor(object):
    def __init__(
        self,
        event_repository: EventRepository,
        new_event_topic: NewEventTopic,
        clock: Clock,
    ) -> None:

        self.event_repository = event_repository
        self.new_event_topic = new_event_topic
        self.clock = clock

    def publish_event(self, request: PublishEventRequest):
        if request.event.synchronous:
            raise CannotProcessSyncEvent()

        if request.external_code is not None:
            publish_detail = self.event_repository.get_by_external_code(external_code=request.external_code)
            if publish_detail is not None:
                return publish_detail

        with tracer.start_as_current_span(
            name="Publish",
            kind=SpanKind.SERVER,
        ) as span:  # type: Span
            if request.async_target is None:
                request.event.synchronous = True

            async_target = None
            if request.target is None and request.async_target is not None:
                async_target = AsyncTarget(
                    reply_subject=request.async_target.reply_subject,
                    id_field=request.async_target.id_field,
                )

            span.set_attribute(key="subject", value=request.event.subject)
            if request.target is not None:
                span.set_attribute(key="type", value="TargetedEvent")
                span.set_attribute(key="target", value=request.target)
            elif async_target is not None:
                span.set_attribute(key="type", value="AsyncTargetEvent")
                span.set_attribute(key="replySubject", value=async_target.reply_subject)
            else:
                span.set_attribute(key="type", value="SyncEvent")

            now = self.clock.utc_now()
            publish_detail = EventPublishDetail(
                id=None,
                created_at=now,
                external_code=request.external_code,
                event=request.event,
                target=request.target,
                async_target=async_target,
                next_retry_at=now,
                fifo=request.fifo,
            )
            span.set_attribute("definitionId", str(publish_detail.id))
            self.new_event_topic.notify(publish_detail=publish_detail)
            return publish_detail



# rfc.py - functions for handling Mexican tax numbers
# coding: utf-8
#
# Copyright (C) 2015 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""RFC (Registro Federal de Contribuyentes, Mexican tax number).

This number is used to identify individuals and companies for tax purposes.

The company number is 12 digits where the first 3 letters or digits are
derived from the name of the company, the following 6 contain the date of
incorporation, followed by 3 check digits.

Personal numbers consist of 13 digits where the first 4 characters from the
person's name, followed by their birth date and 3 check digits.

The first two check digits are calculated based on the person's or company's
full name. The last check digit is calculated over all the preceding digits
in the number. However, it seems a lot of numbers (estimated at around 1.5%
of all numbers) are in use with invalid check digits so this test is disabled
by default.

More information:

* https://www.infomex.org.mx/jspsi/documentos/2005/seguimiento/06101/0610100162005_065.doc
* https://es.wikipedia.org/wiki/Registro_Federal_de_Contribuyentes_(M%C3%A9xico)

An online validation service is available at:

* https://portalsat.plataforma.sat.gob.mx/ConsultaRFC/

>>> validate('GODE 561231 GR8')  # personal number
'GODE561231GR8'
>>> validate('MAB-930714-8T4')  # company number
'MAB9307148T4'
>>> validate('COMG-600703')  # personal number without serial
'COMG600703'
>>> validate('VACE-460910-SX6')
'VACE460910SX6'
>>> validate('VACE-460910-SX6', validate_check_digits=True)
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> format('GODE561231GR8')
'GODE 561231 GR8'
"""

import datetime
import re

from stdnum.exceptions import *
from stdnum.util import clean, to_unicode


# these values should not appear as first part of a personal number
_name_blacklist = set([
    'BUEI', 'BUEY', 'CACA', 'CACO', 'CAGA', 'CAGO', 'CAKA', 'CAKO', 'COGE',
    'COJA', 'COJE', 'COJI', 'COJO', 'CULO', 'FETO', 'GUEY', 'JOTO', 'KACA',
    'KACO', 'KAGA', 'KAGO', 'KAKA', 'KOGE', 'KOJO', 'KULO', 'MAME', 'MAMO',
    'MEAR', 'MEAS', 'MEON', 'MION', 'MOCO', 'MULA', 'PEDA', 'PEDO', 'PENE',
    'PUTA', 'PUTO', 'QULO', 'RATA', 'RUIN',
])


# characters used for checksum calculation,
_alphabet = u'0123456789ABCDEFGHIJKLMN&OPQRSTUVWXYZ Ñ'


def compact(number):
    """Convert the number to the minimal representation. This strips
    surrounding whitespace and separation dash."""
    return clean(number, '-_ ').upper().strip()


def _get_date(number):
    """Convert the part of the number that represents a date into a
    datetime. Note that the century may be incorrect."""
    year = int(number[0:2])
    month = int(number[2:4])
    day = int(number[4:6])
    try:
        return datetime.date(year + 2000, month, day)
    except ValueError:
        raise InvalidComponent()


def calc_check_digit(number):
    """Calculate the check digit. The number passed should not have the
    check digit included."""
    number = to_unicode(number)
    number = ('   ' + number)[-12:]
    check = sum(_alphabet.index(n) * (13 - i) for i, n in enumerate(number))
    return _alphabet[(11 - check) % 11]


def validate(number, validate_check_digits=False):
    """Check if the number is a valid RFC."""
    number = compact(number)
    n = to_unicode(number)
    if len(n) in (10, 13):
        # number assigned to person
        if not re.match(u'^[A-Z&Ñ]{4}[0-9]{6}[0-9A-Z]{0,3}$', n):
            raise InvalidFormat()
        if n[:4] in _name_blacklist:
            raise InvalidComponent()
        _get_date(n[4:10])
    elif len(n) == 12:
        # number assigned to company
        if not re.match(u'^[A-Z&Ñ]{3}[0-9]{6}[0-9A-Z]{3}$', n):
            raise InvalidFormat()
        _get_date(n[3:9])
    else:
        raise InvalidLength()
    if validate_check_digits and len(n) >= 12:
        if not re.match(u'^[1-9A-V][1-9A-Z][0-9A]$', n[-3:]):
            raise InvalidComponent()
        if n[-1] != calc_check_digit(n[:-1]):
            raise InvalidChecksum()
    return number


def is_valid(number, validate_check_digits=False):
    """Check if the number provided is a valid RFC."""
    try:
        return bool(validate(number, validate_check_digits))
    except ValidationError:
        return False


def format(number, separator=' '):
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    if len(number) == 12:
        return separator.join((
            number[:3], number[3:9], number[9:])).strip(separator)
    return separator.join((
        number[:4], number[4:10], number[10:])).strip(separator)



# ext/asyncio/base.py
# Copyright (C) 2020-2023 the SQLAlchemy authors and contributors
# <see AUTHORS file>
#
# This module is part of SQLAlchemy and is released under
# the MIT License: https://www.opensource.org/licenses/mit-license.php

from __future__ import annotations

import abc
import functools
from typing import Any
from typing import AsyncGenerator
from typing import AsyncIterator
from typing import Awaitable
from typing import Callable
from typing import ClassVar
from typing import Dict
from typing import Generator
from typing import Generic
from typing import NoReturn
from typing import Optional
from typing import overload
from typing import Tuple
from typing import TypeVar
import weakref

from . import exc as async_exc
from ... import util
from ...util.typing import Literal
from ...util.typing import Self

_T = TypeVar("_T", bound=Any)
_T_co = TypeVar("_T_co", bound=Any, covariant=True)


_PT = TypeVar("_PT", bound=Any)


class ReversibleProxy(Generic[_PT]):
    _proxy_objects: ClassVar[
        Dict[weakref.ref[Any], weakref.ref[ReversibleProxy[Any]]]
    ] = {}
    __slots__ = ("__weakref__",)

    @overload
    def _assign_proxied(self, target: _PT) -> _PT:
        ...

    @overload
    def _assign_proxied(self, target: None) -> None:
        ...

    def _assign_proxied(self, target: Optional[_PT]) -> Optional[_PT]:
        if target is not None:
            target_ref: weakref.ref[_PT] = weakref.ref(
                target, ReversibleProxy._target_gced
            )
            proxy_ref = weakref.ref(
                self,
                functools.partial(ReversibleProxy._target_gced, target_ref),
            )
            ReversibleProxy._proxy_objects[target_ref] = proxy_ref

        return target

    @classmethod
    def _target_gced(
        cls,
        ref: weakref.ref[_PT],
        proxy_ref: Optional[weakref.ref[Self]] = None,  # noqa: U100
    ) -> None:
        cls._proxy_objects.pop(ref, None)

    @classmethod
    def _regenerate_proxy_for_target(cls, target: _PT) -> Self:
        raise NotImplementedError()

    @overload
    @classmethod
    def _retrieve_proxy_for_target(
        cls,
        target: _PT,
        regenerate: Literal[True] = ...,
    ) -> Self:
        ...

    @overload
    @classmethod
    def _retrieve_proxy_for_target(
        cls, target: _PT, regenerate: bool = True
    ) -> Optional[Self]:
        ...

    @classmethod
    def _retrieve_proxy_for_target(
        cls, target: _PT, regenerate: bool = True
    ) -> Optional[Self]:
        try:
            proxy_ref = cls._proxy_objects[weakref.ref(target)]
        except KeyError:
            pass
        else:
            proxy = proxy_ref()
            if proxy is not None:
                return proxy  # type: ignore

        if regenerate:
            return cls._regenerate_proxy_for_target(target)
        else:
            return None


class StartableContext(Awaitable[_T_co], abc.ABC):
    __slots__ = ()

    @abc.abstractmethod
    async def start(self, is_ctxmanager: bool = False) -> _T_co:
        raise NotImplementedError()

    def __await__(self) -> Generator[Any, Any, _T_co]:
        return self.start().__await__()

    async def __aenter__(self) -> _T_co:
        return await self.start(is_ctxmanager=True)

    @abc.abstractmethod
    async def __aexit__(
        self, type_: Any, value: Any, traceback: Any
    ) -> Optional[bool]:
        pass

    def _raise_for_not_started(self) -> NoReturn:
        raise async_exc.AsyncContextNotStarted(
            "%s context has not been started and object has not been awaited."
            % (self.__class__.__name__)
        )


class GeneratorStartableContext(StartableContext[_T_co]):
    __slots__ = ("gen",)

    gen: AsyncGenerator[_T_co, Any]

    def __init__(
        self,
        func: Callable[..., AsyncIterator[_T_co]],
        args: Tuple[Any, ...],
        kwds: Dict[str, Any],
    ):
        self.gen = func(*args, **kwds)  # type: ignore

    async def start(self, is_ctxmanager: bool = False) -> _T_co:
        try:
            start_value = await util.anext_(self.gen)
        except StopAsyncIteration:
            raise RuntimeError("generator didn't yield") from None

        # if not a context manager, then interrupt the generator, don't
        # let it complete.   this step is technically not needed, as the
        # generator will close in any case at gc time.  not clear if having
        # this here is a good idea or not (though it helps for clarity IMO)
        if not is_ctxmanager:
            await self.gen.aclose()

        return start_value

    async def __aexit__(
        self, typ: Any, value: Any, traceback: Any
    ) -> Optional[bool]:
        # vendored from contextlib.py
        if typ is None:
            try:
                await util.anext_(self.gen)
            except StopAsyncIteration:
                return False
            else:
                raise RuntimeError("generator didn't stop")
        else:
            if value is None:
                # Need to force instantiation so we can reliably
                # tell if we get the same exception back
                value = typ()
            try:
                await util.athrow(self.gen, typ, value, traceback)
            except StopAsyncIteration as exc:
                # Suppress StopIteration *unless* it's the same exception that
                # was passed to throw().  This prevents a StopIteration
                # raised inside the "with" statement from being suppressed.
                return exc is not value
            except RuntimeError as exc:
                # Don't re-raise the passed in exception. (issue27122)
                if exc is value:
                    return False
                # Avoid suppressing if a Stop(Async)Iteration exception
                # was passed to athrow() and later wrapped into a RuntimeError
                # (see PEP 479 for sync generators; async generators also
                # have this behavior). But do this only if the exception
                # wrapped
                # by the RuntimeError is actully Stop(Async)Iteration (see
                # issue29692).
                if (
                    isinstance(value, (StopIteration, StopAsyncIteration))
                    and exc.__cause__ is value
                ):
                    return False
                raise
            except BaseException as exc:
                # only re-raise if it's *not* the exception that was
                # passed to throw(), because __exit__() must not raise
                # an exception unless __exit__() itself failed.  But throw()
                # has to raise the exception to signal propagation, so this
                # fixes the impedance mismatch between the throw() protocol
                # and the __exit__() protocol.
                if exc is not value:
                    raise
                return False
            raise RuntimeError("generator didn't stop after athrow()")


def asyncstartablecontext(
    func: Callable[..., AsyncIterator[_T_co]]
) -> Callable[..., GeneratorStartableContext[_T_co]]:
    """@asyncstartablecontext decorator.

    the decorated function can be called either as ``async with fn()``, **or**
    ``await fn()``.   This is decidedly different from what
    ``@contextlib.asynccontextmanager`` supports, and the usage pattern
    is different as well.

    Typical usage::

        @asyncstartablecontext
        async def some_async_generator(<arguments>):
            <setup>
            try:
                yield <value>
            except GeneratorExit:
                # return value was awaited, no context manager is present
                # and caller will .close() the resource explicitly
                pass
            else:
                <context manager cleanup>


    Above, ``GeneratorExit`` is caught if the function were used as an
    ``await``.  In this case, it's essential that the cleanup does **not**
    occur, so there should not be a ``finally`` block.

    If ``GeneratorExit`` is not invoked, this means we're in ``__aexit__``
    and we were invoked as a context manager, and cleanup should proceed.


    """

    @functools.wraps(func)
    def helper(*args: Any, **kwds: Any) -> GeneratorStartableContext[_T_co]:
        return GeneratorStartableContext(func, args, kwds)

    return helper


class ProxyComparable(ReversibleProxy[_PT]):
    __slots__ = ()

    @util.ro_non_memoized_property
    def _proxied(self) -> _PT:
        raise NotImplementedError()

    def __hash__(self) -> int:
        return id(self)

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, self.__class__)
            and self._proxied == other._proxied
        )

    def __ne__(self, other: Any) -> bool:
        return (
            not isinstance(other, self.__class__)
            or self._proxied != other._proxied
        )



from messagebus import MessageBus
from order_api import OrderService
from typing import Optional

message_bus = None  # type: Optional[MessageBus]


def create_order_service(
    pos_id,     # type: int
):
    return OrderService(
        message_bus=message_bus,
        pos_id=pos_id,
    )



from setuptools import (
    setup,
    find_packages,
)


setup(
    name="NReportsApi",
    version="1.2.0",
    packages=find_packages(),
    install_requires=[],
    url="https://www.e-deploy.com.br",
    author="Lucas de Lima",
    author_email="lucas.lima@e-deploy.com.br"
)



# -*- coding: utf-8 -*-

import json

from bustoken import TK_SLCTRL_PMGR_RUPTURE_UPDATE_ITEMS
from msgbus import (
    FM_PARAM,
    TK_SYS_NAK,
)
from sysactions import (
    StopAction,
    action,
    show_messagebox,
)

from model.customexception import UpdateRuptureException
from .rupture_common import get_rupture_user_id_with_validation
from .. import (
    logger,
    mb_context,
)


@action
def save_rupture(
    pos_id,             # type: str
    enabled_parts,      # type: str
    disabled_parts,     # type: str
):
    # type: (...) -> None

    try:
        enabled_codes_json = json.loads(enabled_parts)
        disabled_codes_json = json.loads(disabled_parts)

        user_id = get_rupture_user_id_with_validation(pos_id=pos_id)
        rupture_status_json = json.dumps({
            "user_id": user_id,
            "enabled": list(map(lambda item: item["product_code"], enabled_codes_json)),
            "disabled": list(map(lambda item: item["product_code"], disabled_codes_json)),
        })
        data = rupture_status_json
        msg = mb_context.MB_EasySendMessage("ORDERMGR", TK_SLCTRL_PMGR_RUPTURE_UPDATE_ITEMS, FM_PARAM, data)
        if msg.token == TK_SYS_NAK:
            raise UpdateRuptureException(msg.data)

        show_messagebox(pos_id, "$RUPTURE_UPDATE_SUCCESS")
    except StopAction:
        pass
    except (Exception, UpdateRuptureException) as e:
        logger.exception('[save_rupture] Exception {}'.format(e))
        show_messagebox(pos_id, "$ERROR_GETTING_DATA", buttons="$EXIT")



import threading
from datetime import timedelta

from opentelemetry import trace
from model import EventRepositoryFactory
from timeutil import (
    Clock,
    Sleeper,
)

tracer = trace.get_tracer(__name__)


class MaintenanceService(object):
    def __init__(
        self,
        event_repository_factory: EventRepositoryFactory,
        sleeper: Sleeper,
        clock: Clock,
        maintenance_interval: int,
        acked_events_ttl: int,
        abandoned_events_ttl: int,
    ):
        self.event_repository_factory = event_repository_factory
        self.sleeper = sleeper
        self.clock = clock
        self.maintenance_interval = float(maintenance_interval / 1000)
        self.acked_events_ttl = acked_events_ttl
        self.abandoned_events_ttl = abandoned_events_ttl

        self.thread = None
        self.running = False

    def start(self):
        if self.thread is None:
            self.thread = threading.Thread(target=self._cleanup, args=[self.maintenance_interval])
            self.thread.daemon = True
            self.running = True
            self.thread.start()

    def stop(self):
        self.running = False

    def cleanup(self):
        with tracer.start_span(name="MaintenanceCleanup") as span:
            span.set_attribute(
                key="startAt",
                value=self.clock.utc_now().isoformat(),
            )
            with self.event_repository_factory.build() as event_repository:
                abandoned_date_limit = self.clock.utc_now() - timedelta(days=self.abandoned_events_ttl)
                event_repository.cleanup_abandoned_events(min_abandoned_date=abandoned_date_limit)

                acked_date_limit = self.clock.utc_now() - timedelta(days=self.acked_events_ttl)
                event_repository.cleanup_acked_events(min_acked_date=acked_date_limit)
            span.set_attribute(
                key="finishAt",
                value=self.clock.utc_now().isoformat(),
            )

    def _cleanup(self, interval):
        while self.running:
            self.cleanup()
            self.sleeper.sleep(interval)



# exceptions.py - collection of stdnum exceptions
# coding: utf-8
#
# Copyright (C) 2013-2022 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Collection of exceptions.

The validation functions of stdnum should raise one of the below exceptions
when validation of the number fails.
"""


__all__ = ['ValidationError', 'InvalidFormat', 'InvalidChecksum',
           'InvalidLength', 'InvalidComponent']


class ValidationError(ValueError):
    """Top-level error for validating numbers.

    This exception should normally not be raised, only subclasses of this
    exception."""

    def __str__(self):
        """Return the exception message."""
        return ''.join(self.args[:1]) or getattr(self, 'message', '')


class InvalidFormat(ValidationError):  # noqa N818
    """Something is wrong with the format of the number.

    This generally means characters or delimiters that are not allowed are
    part of the number or required parts are missing."""

    message = 'The number has an invalid format.'


class InvalidChecksum(ValidationError):  # noqa N818
    """The number's internal checksum or check digit does not match."""

    message = "The number's checksum or check digit is invalid."


class InvalidLength(InvalidFormat):  # noqa N818
    """The length of the number is wrong."""

    message = 'The number has an invalid length.'


class InvalidComponent(ValidationError):  # noqa N818
    """One of the parts of the number has an invalid reference.

    Some part of the number refers to some external entity like a country
    code, a date or a predefined collection of values. The number contains
    some invalid reference."""

    message = 'One of the parts of the number are invalid or unknown.'



from decimal import Decimal

from msgbus import MBEasyContext
from old_helper import BaseRepository
from persistence import Connection
from sysactions import get_poslist
from typing import Tuple


class DrawerRepository(BaseRepository):
    def __init__(
            self,
            mb_context,     # type: MBEasyContext
    ):
        # type: (...) -> None

        super(DrawerRepository, self).__init__(mb_context)
        self.pos_list = get_poslist()

    def get_user_sum_cash_values(
            self,
            period,     # type: str
            session,    # type: str
    ):
        # type: (...) -> Tuple[Decimal, Decimal]

        cash_gross = []
        cash_refund = []

        def inner_func(
                conn    # type: Connection
        ):
            # type: (...) -> None

            query = """
                        SELECT
                            tdsub(
                                    tdsum(DA.SaleCashAmount), 
                                    COALESCE(tdsum(DA.SaleCashChangeAmount), 0)
                                ) AS CashGrossAmount,
                            tdsum(DA.RefundCashAmount) AS RefundCashAmount
                        FROM (
                            SELECT
                                tdsum(
                                    CASE WHEN (Orders.OrderType=0 AND OrderTender.TenderId=0) 
                                        THEN OrderTender.ChangeAmount 
                                        ELSE '0.00' END
                                    ) AS SaleCashChangeAmount,
                                tdsum(
                                    CASE WHEN (Orders.OrderType=0 AND OrderTender.TenderId=0) 
                                        THEN OrderTender.TenderAmount 
                                        ELSE '0.00' END
                                    ) AS SaleCashAmount,
                                COALESCE(
                                    tdsum(
                                        CASE WHEN (OrderTender.TenderId=0 AND Orders.OrderType=1) 
                                            THEN tdsub(OrderTender.TenderAmount,COALESCE(OrderTender.ChangeAmount,'0')) 
                                            ELSE '0.00' END
                                        ),'0.00') AS RefundCashAmount
                            FROM orderdb.Orders Orders
                            JOIN orderdb.OrderTender OrderTender ON OrderTender.OrderId=Orders.OrderId
                            WHERE Orders.StateId=5 AND Orders.OrderType IN (0,1)
                            AND substr(Orders.SessionID, instr(Orders.SessionID, 'period=')+7) = '{}'
                            AND Orders.SessionID = '{}'
                            GROUP BY Orders.BusinessPeriod, Orders.SessionId, Orders.OrderId
                        ) AS DA
                    """.format(period, session)

            for x in conn.select(query):
                cash_gross.append(Decimal(x.get_entry("CashGrossAmount") or 0))
                cash_refund.append(Decimal(x.get_entry("RefundCashAmount") or 0))

        self.execute_parallel_in_all_databases(inner_func, self.pos_list)

        return sum(cash_gross), sum(cash_refund)



# cuit.py - functions for handling Argentinian VAT numbers
# coding: utf-8
#
# Copyright (C) 2009 Mariano Reingart
# Copyright (C) 2011 Sebastián Marró
# Copyright (C) 2008-2011 Cédric Krier
# Copyright (C) 2008-2011 B2CK
# Copyright (C) 2015 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""CUIT (Código Único de Identificación Tributaria, Argentinian tax number).

The CUIT is a taxpayer identification number used for VAT (IVA, Impuesto al
Valor Agregado) and other taxes.

More information:

* https://es.wikipedia.org/wiki/Clave_Única_de_Identificación_Tributaria

>>> validate('20-05536168-2')
'20055361682'
>>> validate('20267565392')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> format('20267565393')
'20-26756539-3'
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' -').strip()


def calc_check_digit(number):
    """Calculate the check digit."""
    weights = (5, 4, 3, 2, 7, 6, 5, 4, 3, 2)
    check = sum(w * int(n) for w, n in zip(weights, number)) % 11
    return '012345678990'[11 - check]


# The different types of CUIT that are known
_cuit_tpes = (
    '20', '23', '24', '27',     # individuals
    '30', '33', '34',           # companies
    '50', '51', '55',           # international purposes
)


def validate(number):
    """Check if the number is a valid CUIT."""
    number = compact(number)
    if len(number) != 11:
        raise InvalidLength()
    if not isdigits(number):
        raise InvalidFormat()
    if number[:2] not in _cuit_tpes:
        raise InvalidComponent()
    if calc_check_digit(number[:-1]) != number[-1]:
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number is a valid CUIT."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    return (number[0:2] + '-' + number[2:10] + '-' + number[10:])



from typing import Any, Optional, List
from messagebus import MessageBus, Message, Event

from ._MessageProcessor import MessageProcessor
from ._MessageProcessorCallback import MessageProcessorCallback


class DefaultMessageProcessorExecutor(object):
    def __init__(self, event_processor, callbacks=None):
        # type: (MessageProcessor, Optional[List[MessageProcessorCallback]]) -> None
        self.event_processor = event_processor
        self.callbacks = callbacks

    def execute_processor(self, message_bus, message, event, data):
        # type: (MessageBus, Optional[Message], Optional[Event], Any) -> None
        event_name = self.event_processor.get_processor_name()
        request_id = self.event_processor.generate_unique_id()

        self._process_begun(event_name, request_id, data)
        input_obj = None
        try:
            event_type = None
            if event is not None and event.event_type is not None:
                event_type = event.event_type
            input_obj = self.event_processor.parse_data(data, event_type)
            event_name = self.event_processor.get_processor_name()
            self._input_parsed(event_name, request_id, input_obj)

            output_obj = self.event_processor.call_business(input_obj)
            self._business_called(event_name, request_id, input_obj, output_obj)
            self.event_processor.format_response(message_bus, message, event, input_obj, output_obj)
            self._process_finished(event_name, request_id, input_obj, output_obj)
        except BaseException as ex:
            if input_obj is not None:
                self.event_processor.format_exception(message_bus, message, event, input_obj, ex)
                self._unhandled_exception(event_name, request_id, input_obj, ex)
            else:
                self.event_processor.format_parse_exception(message_bus, message, event, data, ex)
                self._parse_exception(event_name, request_id, data, ex)

    def _process_begun(self, event_name, request_id, data):
        # type: (str, str, bytes) -> None
        if self.callbacks is not None:
            for callback in self.callbacks:
                callback.process_begun(event_name, request_id, data)

    def _input_parsed(self, event_name, request_id, input_obj):
        # type: (str, str, Any) -> None
        if self.callbacks is not None:
            for callback in self.callbacks:
                callback.input_parsed(event_name, request_id, input_obj)

    def _business_called(self, event_name, request_id, input_obj, output_obj):
        # type: (str, str, Any, Any) -> None
        if self.callbacks is not None:
            for callback in self.callbacks:
                callback.business_called(event_name, request_id, input_obj, output_obj)

    def _process_finished(self, event_name, request_id, input_obj, output_obj):
        # type: (str, str, Any, Any) -> None
        if self.callbacks is not None:
            for callback in self.callbacks:
                callback.process_finished(event_name, request_id, input_obj, output_obj)

    def _parse_exception(self, event_name, request_id, data, exception):
        # type: (str, str, bytes, BaseException) -> None
        if self.callbacks is not None:
            for callback in self.callbacks:
                callback.parse_exception(event_name, request_id, data, exception)

    def _unhandled_exception(self, event_name, request_id, input_obj, exception):
        # type: (str, str, Any, BaseException) -> None
        if self.callbacks is not None:
            for callback in self.callbacks:
                callback.unhandled_exception(event_name, request_id, input_obj, exception)



# -*- coding: utf-8 -*-
# vi:tabstop=4:expandtab:sw=4
"""Transliterate Unicode text into plain 7-bit ASCII.

Example usage:
>>> from unidecode import unidecode
>>> unidecode(u"\u5317\u4EB0")
"Bei Jing "

The transliteration uses a straightforward map, and doesn't have alternatives
for the same character based on language, position, or anything else.

In Python 3, a standard string object will be returned. If you need bytes, use:
>>> unidecode("Κνωσός").encode("ascii")
b'Knosos'
"""
import warnings
from sys import version_info

Cache = {}

class UnidecodeError(ValueError):
    def __init__(self, message, index=None):
        """Raised for Unidecode-related errors.

        The index attribute contains the index of the character that caused
        the error.
        """
        super(UnidecodeError, self).__init__(message)
        self.index = index

def _warn_if_not_unicode(string):
    if version_info[0] < 3 and not isinstance(string, unicode):
        warnings.warn(  "Argument %r is not an unicode object. "
                        "Passing an encoded string will likely have "
                        "unexpected results." % (type(string),),
                        RuntimeWarning, 2)


def unidecode_expect_ascii(string, errors='ignore', replace_str='?'):
    """Transliterate an Unicode object into an ASCII string

    >>> unidecode(u"\u5317\u4EB0")
    "Bei Jing "

    This function first tries to convert the string using ASCII codec.
    If it fails (because of non-ASCII characters), it falls back to
    transliteration using the character tables.

    This is approx. five times faster if the string only contains ASCII
    characters, but slightly slower than unicode_expect_nonascii if
    non-ASCII characters are present.

    errors specifies what to do with characters that have not been
    found in replacement tables. The default is 'ignore' which ignores
    the character. 'strict' raises an UnidecodeError. 'replace'
    substitutes the character with replace_str (default is '?').
    'preserve' keeps the original character.

    Note that if 'preserve' is used the returned string might not be
    ASCII!
    """

    _warn_if_not_unicode(string)
    try:
        bytestring = string.encode('ASCII')
    except UnicodeEncodeError:
        pass
    else:
        if version_info[0] >= 3:
            return string
        else:
            return bytestring

    return _unidecode(string, errors, replace_str)

def unidecode_expect_nonascii(string, errors='ignore', replace_str='?'):
    """Transliterate an Unicode object into an ASCII string

    >>> unidecode(u"\u5317\u4EB0")
    "Bei Jing "

    See unidecode_expect_ascii.
    """

    _warn_if_not_unicode(string)
    return _unidecode(string, errors, replace_str)

unidecode = unidecode_expect_ascii

def _get_repl_str(char):
    codepoint = ord(char)

    if codepoint < 0x80:
        # Already ASCII
        return str(char)

    if codepoint > 0xeffff:
        # No data on characters in Private Use Area and above.
        return None

    if 0xd800 <= codepoint <= 0xdfff:
        warnings.warn(  "Surrogate character %r will be ignored. "
                        "You might be using a narrow Python build." % (char,),
                        RuntimeWarning, 2)

    section = codepoint >> 8   # Chop off the last two hex digits
    position = codepoint % 256 # Last two hex digits

    try:
        table = Cache[section]
    except KeyError:
        try:
            mod = __import__('unidecode.x%03x'%(section), globals(), locals(), ['data'])
        except ImportError:
            # No data on this character
            Cache[section] = None
            return None

        Cache[section] = table = mod.data

    if table and len(table) > position:
        return table[position]
    else:
        return None

def _unidecode(string, errors, replace_str):
    retval = []

    for index, char in enumerate(string):
        repl = _get_repl_str(char)

        if repl is None:
            if errors == 'ignore':
                repl = ''
            elif errors == 'strict':
                raise UnidecodeError('no replacement found for character %r '
                        'in position %d' % (char, index), index)
            elif errors == 'replace':
                repl = replace_str
            elif errors == 'preserve':
                repl = char
            else:
                raise UnidecodeError('invalid value for errors parameter %r' % (errors,))

        retval.append(repl)

    return ''.join(retval)



from xml.etree import cElementTree as eTree
from datetime import datetime
from typing import (
    Dict,
    Union,
    List, Optional,
)
from ._Table import Table
from ._ServiceTender import ServiceTender
from order_api.dto import Order as OrderApi
from order_api.dto.order_xml_parser import parse_order


def get_table_attribute_value(
    key,                # type: str
    table_xml_attrib,   # type: Dict[str, str]
    default_value,      # type: Union[str, int, None]
):
    # type: (...) -> Union[str, int, None]

    return table_xml_attrib.get(key, default_value)


def parse_tab_names(
    table_xml,     # type: eTree.Element
):
    # type: (...) -> Dict[str, str]

    service_children = table_xml.findall("ServiceChildren/ServiceChild")
    tabs_names = {}
    for service_child in service_children:
        tabs_names.update({service_child.attrib["tableId"]: service_child.attrib["tabId"]})

    return tabs_names


def parse_table_name(
    table_id,      # type: str
    table_xml,     # type: eTree.Element
):
    # type: (...) -> str

    property_xml = table_xml.findall(".//Properties[@TAB_ID]")
    tab_name = property_xml[0].attrib["TAB_ID"] if property_xml else None
    table_name = table_xml.attrib["name"] if "name" in table_xml.attrib else table_id
    return tab_name if tab_name else table_name


def parse_service_tenders(
    service_tenders,  # type: List[eTree.Element]
):
    # type: (...) ->  List[ServiceTender]

    service_tender_parsed = []      # type: List[ServiceTender]
    for service_tender in service_tenders:
        tender_amount = float(service_tender.get('tenderAmount'))
        if tender_amount > 0:
            service_tender_parsed.append(
                ServiceTender(
                    description=service_tender.get('tenderDescr'),
                    amount=tender_amount,
                )
            )

    return service_tender_parsed


def parse_table(
    table_xml,  # type: eTree.Element
):
    # type: (...) ->  Table

    table_id = table_xml.get("id")
    user_id = table_xml.attrib.get("userId", "")
    status = int(table_xml.attrib.get("status")) if table_xml.attrib.get("status") else 0

    table_name = parse_table_name(
        table_id=table_id,
        table_xml=table_xml,
    )
    tabs_names = parse_tab_names(table_xml=table_xml)

    table_mode = None
    if "MODE" in table_xml.find("Properties").attrib:
        table_mode = table_xml.find("Properties").attrib["MODE"]

    total_amount = float(get_table_attribute_value(
        table_xml_attrib=table_xml.attrib,
        key="totalAmount",
        default_value=0,
    ))
    service_seats = int(get_table_attribute_value(
        table_xml_attrib=table_xml.attrib,
        key="serviceSeats",
        default_value=0,
    ))
    total_per_seat = float(get_table_attribute_value(
        table_xml_attrib=table_xml.attrib,
        key="totalPerSeat",
        default_value=0,
    ))
    total_discounts = float(get_table_attribute_value(
        table_xml_attrib=table_xml.attrib,
        key="discountAmount",
        default_value=0,
    ))
    tip = float(table_xml.get("tip") if table_xml.get("tip") else 0.0)
    tip_rate = float(table_xml.get("tipRate") if table_xml.get("tipRate") else 0.0)

    start_date = get_table_attribute_value(
        table_xml_attrib=table_xml.attrib,
        key="startTS",
        default_value=None,
    )
    start_date = datetime.strptime(start_date, "%Y-%m-%d %H:%M:%S") if start_date else None

    service_tenders = table_xml.findall('ServiceTenders/ServiceTender')
    service_tenders = parse_service_tenders(service_tenders=service_tenders)

    orders_xml = table_xml.findall("Orders/Order")
    orders = []     # type: Optional[List[OrderApi]]
    for order_xml in orders_xml:
        order_api = parse_order(order_xml=order_xml)
        orders.append(order_api)

    return Table(
        table_id=table_id,
        user_id=user_id,
        table_name=table_name,
        tabs_names=tabs_names,
        table_mode=table_mode,
        tip_rate=tip_rate,
        tip=tip,
        total_amount=total_amount,
        service_seats=service_seats,
        total_per_seat=total_per_seat,
        total_discounts=total_discounts,
        start_date=start_date,
        service_tenders=service_tenders,
        orders=orders,
        status=status,
    )



# figi.py - functions for handling FIGI numbers
#
# Copyright (C) 2018 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""FIGI (Financial Instrument Global Identifier).

The Financial Instrument Global Identifier (FIGI) is a 12-character
alpha-numerical unique identifier of financial instruments such as common
stock, options, derivatives, futures, corporate and government bonds,
municipals, currencies, and mortgage products.

More information:

* https://openfigi.com/
* https://en.wikipedia.org/wiki/Financial_Instrument_Global_Identifier

>>> validate('BBG000BLNQ16')
'BBG000BLNQ16'
>>> validate('BBG000BLNQ14')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' ').strip().upper()


def calc_check_digit(number):
    """Calculate the check digits for the number."""
    # we use the full alphabet for the check digit calculation
    alphabet = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    # convert to numeric first, then double some, then sum individual digits
    number = ''.join(
        str(alphabet.index(n) * (1, 2)[i % 2])
        for i, n in enumerate(number[:11]))
    return str((10 - sum(int(n) for n in number)) % 10)


def validate(number):
    """Check if the number provided is a valid FIGI."""
    number = compact(number)
    if not all(x in '0123456789BCDFGHJKLMNPQRSTVWXYZ' for x in number):
        raise InvalidFormat()
    if len(number) != 12:
        raise InvalidLength()
    if isdigits(number[0]) or isdigits(number[1]):
        raise InvalidFormat()
    if number[:2] in ('BS', 'BM', 'GG', 'GB', 'VG'):
        raise InvalidComponent()
    if number[2] != 'G':
        raise InvalidComponent()
    if calc_check_digit(number[:-1]) != number[-1]:
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number provided is a valid FIGI."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



import re

from logging import Logger

from mw_helper import ensure_iterable
from production.box import ProductionBox
from production.repository import ProductRepository
from typing import List, Optional, Union

from ._ItemFilterBox import ItemFilterBox


class JitItemFilterBox(ItemFilterBox):
    def __init__(
            self,
            name,                       # type: str
            sons,                       # type: Optional[Union[List[ProductionBox]], ProductionBox]
            repository,                 # type: ProductRepository
            allowed_jit_lines=None,     # type: Optional[Union[str, List[str]]]
            forbidden_jit_lines=None,   # type: Optional[Union[str, List[str]]]
            logger=None,                # type: Logger
    ):
        # type: (...) -> None

        self.allowed_jit_lines = ensure_iterable(allowed_jit_lines) if allowed_jit_lines is not None else None
        self.forbidden_jit_lines = ensure_iterable(forbidden_jit_lines) if forbidden_jit_lines is not None else None

        allowed_codes = None
        if allowed_jit_lines is not None:
            allowed_jit_lines = ensure_iterable(allowed_jit_lines)
            allowed_codes = repository.get_part_codes_of_jit_lines(allowed_jit_lines)

        forbidden_codes = None
        if forbidden_jit_lines is not None:
            forbidden_jit_lines = ensure_iterable(forbidden_jit_lines)
            forbidden_codes = repository.get_part_codes_of_jit_lines(forbidden_jit_lines)

        super(JitItemFilterBox, self).__init__(name, sons, allowed_codes, forbidden_codes, logger)

    def forbidden_part_code(self, item):
        if not self.forbidden_jit_lines:
            return False

        item_jit_lines = re.findall('(\\w+)', item.jit_lines)
        for jit_line in self.forbidden_jit_lines:
            if jit_line in item_jit_lines:
                return True

        return False

    def allowed_part_code(self, item):
        if not self.allowed_jit_lines:
            return True

        item_jit_lines = re.findall('(\\w+)', item.jit_lines)
        for jit_line in self.allowed_jit_lines:
            if jit_line in item_jit_lines:
                return True

        return False



from __future__ import annotations

import collections
import re
from typing import Any
from typing import Callable
from typing import cast
from typing import Collection
from typing import Deque
from typing import Dict
from typing import FrozenSet
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional
from typing import overload
from typing import Sequence
from typing import Set
from typing import Tuple
from typing import TYPE_CHECKING
from typing import TypeVar
from typing import Union

from sqlalchemy import util as sqlautil

from .. import util
from ..util import not_none

if TYPE_CHECKING:
    from typing import Literal

_RevIdType = Union[str, List[str], Tuple[str, ...]]
_GetRevArg = Union[
    str,
    Iterable[Optional[str]],
    Iterable[str],
]
_RevisionIdentifierType = Union[str, Tuple[str, ...], None]
_RevisionOrStr = Union["Revision", str]
_RevisionOrBase = Union["Revision", "Literal['base']"]
_InterimRevisionMapType = Dict[str, "Revision"]
_RevisionMapType = Dict[Union[None, str, Tuple[()]], Optional["Revision"]]
_T = TypeVar("_T")
_TR = TypeVar("_TR", bound=Optional[_RevisionOrStr])

_relative_destination = re.compile(r"(?:(.+?)@)?(\w+)?((?:\+|-)\d+)")
_revision_illegal_chars = ["@", "-", "+"]


class RevisionError(Exception):
    pass


class RangeNotAncestorError(RevisionError):
    def __init__(
        self, lower: _RevisionIdentifierType, upper: _RevisionIdentifierType
    ) -> None:
        self.lower = lower
        self.upper = upper
        super().__init__(
            "Revision %s is not an ancestor of revision %s"
            % (lower or "base", upper or "base")
        )


class MultipleHeads(RevisionError):
    def __init__(self, heads: Sequence[str], argument: Optional[str]) -> None:
        self.heads = heads
        self.argument = argument
        super().__init__(
            "Multiple heads are present for given argument '%s'; "
            "%s" % (argument, ", ".join(heads))
        )


class ResolutionError(RevisionError):
    def __init__(self, message: str, argument: str) -> None:
        super().__init__(message)
        self.argument = argument


class CycleDetected(RevisionError):
    kind = "Cycle"

    def __init__(self, revisions: Sequence[str]) -> None:
        self.revisions = revisions
        super().__init__(
            "%s is detected in revisions (%s)"
            % (self.kind, ", ".join(revisions))
        )


class DependencyCycleDetected(CycleDetected):
    kind = "Dependency cycle"

    def __init__(self, revisions: Sequence[str]) -> None:
        super().__init__(revisions)


class LoopDetected(CycleDetected):
    kind = "Self-loop"

    def __init__(self, revision: str) -> None:
        super().__init__([revision])


class DependencyLoopDetected(DependencyCycleDetected, LoopDetected):
    kind = "Dependency self-loop"

    def __init__(self, revision: Sequence[str]) -> None:
        super().__init__(revision)


class RevisionMap:
    """Maintains a map of :class:`.Revision` objects.

    :class:`.RevisionMap` is used by :class:`.ScriptDirectory` to maintain
    and traverse the collection of :class:`.Script` objects, which are
    themselves instances of :class:`.Revision`.

    """

    def __init__(self, generator: Callable[[], Iterable[Revision]]) -> None:
        """Construct a new :class:`.RevisionMap`.

        :param generator: a zero-arg callable that will generate an iterable
         of :class:`.Revision` instances to be used.   These are typically
         :class:`.Script` subclasses within regular Alembic use.

        """
        self._generator = generator

    @util.memoized_property
    def heads(self) -> Tuple[str, ...]:
        """All "head" revisions as strings.

        This is normally a tuple of length one,
        unless unmerged branches are present.

        :return: a tuple of string revision numbers.

        """
        self._revision_map
        return self.heads

    @util.memoized_property
    def bases(self) -> Tuple[str, ...]:
        """All "base" revisions as strings.

        These are revisions that have a ``down_revision`` of None,
        or empty tuple.

        :return: a tuple of string revision numbers.

        """
        self._revision_map
        return self.bases

    @util.memoized_property
    def _real_heads(self) -> Tuple[str, ...]:
        """All "real" head revisions as strings.

        :return: a tuple of string revision numbers.

        """
        self._revision_map
        return self._real_heads

    @util.memoized_property
    def _real_bases(self) -> Tuple[str, ...]:
        """All "real" base revisions as strings.

        :return: a tuple of string revision numbers.

        """
        self._revision_map
        return self._real_bases

    @util.memoized_property
    def _revision_map(self) -> _RevisionMapType:
        """memoized attribute, initializes the revision map from the
        initial collection.

        """
        # Ordering required for some tests to pass (but not required in
        # general)
        map_: _InterimRevisionMapType = sqlautil.OrderedDict()

        heads: Set[Revision] = sqlautil.OrderedSet()
        _real_heads: Set[Revision] = sqlautil.OrderedSet()
        bases: Tuple[Revision, ...] = ()
        _real_bases: Tuple[Revision, ...] = ()

        has_branch_labels = set()
        all_revisions = set()

        for revision in self._generator():
            all_revisions.add(revision)

            if revision.revision in map_:
                util.warn(
                    "Revision %s is present more than once" % revision.revision
                )
            map_[revision.revision] = revision
            if revision.branch_labels:
                has_branch_labels.add(revision)

            heads.add(revision)
            _real_heads.add(revision)
            if revision.is_base:
                bases += (revision,)
            if revision._is_real_base:
                _real_bases += (revision,)

        # add the branch_labels to the map_.  We'll need these
        # to resolve the dependencies.
        rev_map = map_.copy()
        self._map_branch_labels(
            has_branch_labels, cast(_RevisionMapType, map_)
        )

        # resolve dependency names from branch labels and symbolic
        # names
        self._add_depends_on(all_revisions, cast(_RevisionMapType, map_))

        for rev in map_.values():
            for downrev in rev._all_down_revisions:
                if downrev not in map_:
                    util.warn(
                        "Revision %s referenced from %s is not present"
                        % (downrev, rev)
                    )
                down_revision = map_[downrev]
                down_revision.add_nextrev(rev)
                if downrev in rev._versioned_down_revisions:
                    heads.discard(down_revision)
                _real_heads.discard(down_revision)

        # once the map has downrevisions populated, the dependencies
        # can be further refined to include only those which are not
        # already ancestors
        self._normalize_depends_on(all_revisions, cast(_RevisionMapType, map_))
        self._detect_cycles(rev_map, heads, bases, _real_heads, _real_bases)

        revision_map: _RevisionMapType = dict(map_.items())
        revision_map[None] = revision_map[()] = None
        self.heads = tuple(rev.revision for rev in heads)
        self._real_heads = tuple(rev.revision for rev in _real_heads)
        self.bases = tuple(rev.revision for rev in bases)
        self._real_bases = tuple(rev.revision for rev in _real_bases)

        self._add_branches(has_branch_labels, revision_map)
        return revision_map

    def _detect_cycles(
        self,
        rev_map: _InterimRevisionMapType,
        heads: Set[Revision],
        bases: Tuple[Revision, ...],
        _real_heads: Set[Revision],
        _real_bases: Tuple[Revision, ...],
    ) -> None:
        if not rev_map:
            return
        if not heads or not bases:
            raise CycleDetected(list(rev_map))
        total_space = {
            rev.revision
            for rev in self._iterate_related_revisions(
                lambda r: r._versioned_down_revisions,
                heads,
                map_=cast(_RevisionMapType, rev_map),
            )
        }.intersection(
            rev.revision
            for rev in self._iterate_related_revisions(
                lambda r: r.nextrev,
                bases,
                map_=cast(_RevisionMapType, rev_map),
            )
        )
        deleted_revs = set(rev_map.keys()) - total_space
        if deleted_revs:
            raise CycleDetected(sorted(deleted_revs))

        if not _real_heads or not _real_bases:
            raise DependencyCycleDetected(list(rev_map))
        total_space = {
            rev.revision
            for rev in self._iterate_related_revisions(
                lambda r: r._all_down_revisions,
                _real_heads,
                map_=cast(_RevisionMapType, rev_map),
            )
        }.intersection(
            rev.revision
            for rev in self._iterate_related_revisions(
                lambda r: r._all_nextrev,
                _real_bases,
                map_=cast(_RevisionMapType, rev_map),
            )
        )
        deleted_revs = set(rev_map.keys()) - total_space
        if deleted_revs:
            raise DependencyCycleDetected(sorted(deleted_revs))

    def _map_branch_labels(
        self, revisions: Collection[Revision], map_: _RevisionMapType
    ) -> None:
        for revision in revisions:
            if revision.branch_labels:
                assert revision._orig_branch_labels is not None
                for branch_label in revision._orig_branch_labels:
                    if branch_label in map_:
                        map_rev = map_[branch_label]
                        assert map_rev is not None
                        raise RevisionError(
                            "Branch name '%s' in revision %s already "
                            "used by revision %s"
                            % (
                                branch_label,
                                revision.revision,
                                map_rev.revision,
                            )
                        )
                    map_[branch_label] = revision

    def _add_branches(
        self, revisions: Collection[Revision], map_: _RevisionMapType
    ) -> None:
        for revision in revisions:
            if revision.branch_labels:
                revision.branch_labels.update(revision.branch_labels)
                for node in self._get_descendant_nodes(
                    [revision], map_, include_dependencies=False
                ):
                    node.branch_labels.update(revision.branch_labels)

                parent = node
                while (
                    parent
                    and not parent._is_real_branch_point
                    and not parent.is_merge_point
                ):
                    parent.branch_labels.update(revision.branch_labels)
                    if parent.down_revision:
                        parent = map_[parent.down_revision]
                    else:
                        break

    def _add_depends_on(
        self, revisions: Collection[Revision], map_: _RevisionMapType
    ) -> None:
        """Resolve the 'dependencies' for each revision in a collection
        in terms of actual revision ids, as opposed to branch labels or other
        symbolic names.

        The collection is then assigned to the _resolved_dependencies
        attribute on each revision object.

        """

        for revision in revisions:
            if revision.dependencies:
                deps = [
                    map_[dep] for dep in util.to_tuple(revision.dependencies)
                ]
                revision._resolved_dependencies = tuple(
                    [d.revision for d in deps if d is not None]
                )
            else:
                revision._resolved_dependencies = ()

    def _normalize_depends_on(
        self, revisions: Collection[Revision], map_: _RevisionMapType
    ) -> None:
        """Create a collection of "dependencies" that omits dependencies
        that are already ancestor nodes for each revision in a given
        collection.

        This builds upon the _resolved_dependencies collection created in the
        _add_depends_on() method, looking in the fully populated revision map
        for ancestors, and omitting them as the _resolved_dependencies
        collection as it is copied to a new collection. The new collection is
        then assigned to the _normalized_resolved_dependencies attribute on
        each revision object.

        The collection is then used to determine the immediate "down revision"
        identifiers for this revision.

        """

        for revision in revisions:
            if revision._resolved_dependencies:
                normalized_resolved = set(revision._resolved_dependencies)
                for rev in self._get_ancestor_nodes(
                    [revision],
                    include_dependencies=False,
                    map_=cast(_RevisionMapType, map_),
                ):
                    if rev is revision:
                        continue
                    elif rev._resolved_dependencies:
                        normalized_resolved.difference_update(
                            rev._resolved_dependencies
                        )

                revision._normalized_resolved_dependencies = tuple(
                    normalized_resolved
                )
            else:
                revision._normalized_resolved_dependencies = ()

    def add_revision(self, revision: Revision, _replace: bool = False) -> None:
        """add a single revision to an existing map.

        This method is for single-revision use cases, it's not
        appropriate for fully populating an entire revision map.

        """
        map_ = self._revision_map
        if not _replace and revision.revision in map_:
            util.warn(
                "Revision %s is present more than once" % revision.revision
            )
        elif _replace and revision.revision not in map_:
            raise Exception("revision %s not in map" % revision.revision)

        map_[revision.revision] = revision

        revisions = [revision]
        self._add_branches(revisions, map_)
        self._map_branch_labels(revisions, map_)
        self._add_depends_on(revisions, map_)

        if revision.is_base:
            self.bases += (revision.revision,)
        if revision._is_real_base:
            self._real_bases += (revision.revision,)

        for downrev in revision._all_down_revisions:
            if downrev not in map_:
                util.warn(
                    "Revision %s referenced from %s is not present"
                    % (downrev, revision)
                )
            not_none(map_[downrev]).add_nextrev(revision)

        self._normalize_depends_on(revisions, map_)

        if revision._is_real_head:
            self._real_heads = tuple(
                head
                for head in self._real_heads
                if head
                not in set(revision._all_down_revisions).union(
                    [revision.revision]
                )
            ) + (revision.revision,)
        if revision.is_head:
            self.heads = tuple(
                head
                for head in self.heads
                if head
                not in set(revision._versioned_down_revisions).union(
                    [revision.revision]
                )
            ) + (revision.revision,)

    def get_current_head(
        self, branch_label: Optional[str] = None
    ) -> Optional[str]:
        """Return the current head revision.

        If the script directory has multiple heads
        due to branching, an error is raised;
        :meth:`.ScriptDirectory.get_heads` should be
        preferred.

        :param branch_label: optional branch name which will limit the
         heads considered to those which include that branch_label.

        :return: a string revision number.

        .. seealso::

            :meth:`.ScriptDirectory.get_heads`

        """
        current_heads: Sequence[str] = self.heads
        if branch_label:
            current_heads = self.filter_for_lineage(
                current_heads, branch_label
            )
        if len(current_heads) > 1:
            raise MultipleHeads(
                current_heads,
                "%s@head" % branch_label if branch_label else "head",
            )

        if current_heads:
            return current_heads[0]
        else:
            return None

    def _get_base_revisions(self, identifier: str) -> Tuple[str, ...]:
        return self.filter_for_lineage(self.bases, identifier)

    def get_revisions(
        self, id_: Optional[_GetRevArg]
    ) -> Tuple[Optional[_RevisionOrBase], ...]:
        """Return the :class:`.Revision` instances with the given rev id
        or identifiers.

        May be given a single identifier, a sequence of identifiers, or the
        special symbols "head" or "base".  The result is a tuple of one
        or more identifiers, or an empty tuple in the case of "base".

        In the cases where 'head', 'heads' is requested and the
        revision map is empty, returns an empty tuple.

        Supports partial identifiers, where the given identifier
        is matched against all identifiers that start with the given
        characters; if there is exactly one match, that determines the
        full revision.

        """

        if isinstance(id_, (list, tuple, set, frozenset)):
            return sum([self.get_revisions(id_elem) for id_elem in id_], ())
        else:
            resolved_id, branch_label = self._resolve_revision_number(id_)
            if len(resolved_id) == 1:
                try:
                    rint = int(resolved_id[0])
                    if rint < 0:
                        # branch@-n -> walk down from heads
                        select_heads = self.get_revisions("heads")
                        if branch_label is not None:
                            select_heads = tuple(
                                head
                                for head in select_heads
                                if branch_label
                                in is_revision(head).branch_labels
                            )
                        return tuple(
                            self._walk(head, steps=rint)
                            for head in select_heads
                        )
                except ValueError:
                    # couldn't resolve as integer
                    pass
            return tuple(
                self._revision_for_ident(rev_id, branch_label)
                for rev_id in resolved_id
            )

    def get_revision(self, id_: Optional[str]) -> Optional[Revision]:
        """Return the :class:`.Revision` instance with the given rev id.

        If a symbolic name such as "head" or "base" is given, resolves
        the identifier into the current head or base revision.  If the symbolic
        name refers to multiples, :class:`.MultipleHeads` is raised.

        Supports partial identifiers, where the given identifier
        is matched against all identifiers that start with the given
        characters; if there is exactly one match, that determines the
        full revision.

        """

        resolved_id, branch_label = self._resolve_revision_number(id_)
        if len(resolved_id) > 1:
            raise MultipleHeads(resolved_id, id_)

        resolved: Union[str, Tuple[()]] = resolved_id[0] if resolved_id else ()
        return self._revision_for_ident(resolved, branch_label)

    def _resolve_branch(self, branch_label: str) -> Optional[Revision]:
        try:
            branch_rev = self._revision_map[branch_label]
        except KeyError:
            try:
                nonbranch_rev = self._revision_for_ident(branch_label)
            except ResolutionError as re:
                raise ResolutionError(
                    "No such branch: '%s'" % branch_label, branch_label
                ) from re

            else:
                return nonbranch_rev
        else:
            return branch_rev

    def _revision_for_ident(
        self,
        resolved_id: Union[str, Tuple[()], None],
        check_branch: Optional[str] = None,
    ) -> Optional[Revision]:
        branch_rev: Optional[Revision]
        if check_branch:
            branch_rev = self._resolve_branch(check_branch)
        else:
            branch_rev = None

        revision: Union[Optional[Revision], Literal[False]]
        try:
            revision = self._revision_map[resolved_id]
        except KeyError:
            # break out to avoid misleading py3k stack traces
            revision = False
        revs: Sequence[str]
        if revision is False:
            assert resolved_id
            # do a partial lookup
            revs = [
                x
                for x in self._revision_map
                if x and len(x) > 3 and x.startswith(resolved_id)
            ]

            if branch_rev:
                revs = self.filter_for_lineage(revs, check_branch)
            if not revs:
                raise ResolutionError(
                    "No such revision or branch '%s'%s"
                    % (
                        resolved_id,
                        (
                            "; please ensure at least four characters are "
                            "present for partial revision identifier matches"
                            if len(resolved_id) < 4
                            else ""
                        ),
                    ),
                    resolved_id,
                )
            elif len(revs) > 1:
                raise ResolutionError(
                    "Multiple revisions start "
                    "with '%s': %s..."
                    % (resolved_id, ", ".join("'%s'" % r for r in revs[0:3])),
                    resolved_id,
                )
            else:
                revision = self._revision_map[revs[0]]

        if check_branch and revision is not None:
            assert branch_rev is not None
            assert resolved_id
            if not self._shares_lineage(
                revision.revision, branch_rev.revision
            ):
                raise ResolutionError(
                    "Revision %s is not a member of branch '%s'"
                    % (revision.revision, check_branch),
                    resolved_id,
                )
        return revision

    def _filter_into_branch_heads(
        self, targets: Iterable[Optional[_RevisionOrBase]]
    ) -> Set[Optional[_RevisionOrBase]]:
        targets = set(targets)

        for rev in list(targets):
            assert rev
            if targets.intersection(
                self._get_descendant_nodes([rev], include_dependencies=False)
            ).difference([rev]):
                targets.discard(rev)
        return targets

    def filter_for_lineage(
        self,
        targets: Iterable[_TR],
        check_against: Optional[str],
        include_dependencies: bool = False,
    ) -> Tuple[_TR, ...]:
        id_, branch_label = self._resolve_revision_number(check_against)

        shares = []
        if branch_label:
            shares.append(branch_label)
        if id_:
            shares.extend(id_)

        return tuple(
            tg
            for tg in targets
            if self._shares_lineage(
                tg, shares, include_dependencies=include_dependencies
            )
        )

    def _shares_lineage(
        self,
        target: Optional[_RevisionOrStr],
        test_against_revs: Sequence[_RevisionOrStr],
        include_dependencies: bool = False,
    ) -> bool:
        if not test_against_revs:
            return True
        if not isinstance(target, Revision):
            resolved_target = not_none(self._revision_for_ident(target))
        else:
            resolved_target = target

        resolved_test_against_revs = [
            self._revision_for_ident(test_against_rev)
            if not isinstance(test_against_rev, Revision)
            else test_against_rev
            for test_against_rev in util.to_tuple(
                test_against_revs, default=()
            )
        ]

        return bool(
            set(
                self._get_descendant_nodes(
                    [resolved_target],
                    include_dependencies=include_dependencies,
                )
            )
            .union(
                self._get_ancestor_nodes(
                    [resolved_target],
                    include_dependencies=include_dependencies,
                )
            )
            .intersection(resolved_test_against_revs)
        )

    def _resolve_revision_number(
        self, id_: Optional[_GetRevArg]
    ) -> Tuple[Tuple[str, ...], Optional[str]]:
        branch_label: Optional[str]
        if isinstance(id_, str) and "@" in id_:
            branch_label, id_ = id_.split("@", 1)

        elif id_ is not None and (
            (isinstance(id_, tuple) and id_ and not isinstance(id_[0], str))
            or not isinstance(id_, (str, tuple))
        ):
            raise RevisionError(
                "revision identifier %r is not a string; ensure database "
                "driver settings are correct" % (id_,)
            )

        else:
            branch_label = None

        # ensure map is loaded
        self._revision_map
        if id_ == "heads":
            if branch_label:
                return (
                    self.filter_for_lineage(self.heads, branch_label),
                    branch_label,
                )
            else:
                return self._real_heads, branch_label
        elif id_ == "head":
            current_head = self.get_current_head(branch_label)
            if current_head:
                return (current_head,), branch_label
            else:
                return (), branch_label
        elif id_ == "base" or id_ is None:
            return (), branch_label
        else:
            return util.to_tuple(id_, default=None), branch_label

    def iterate_revisions(
        self,
        upper: _RevisionIdentifierType,
        lower: _RevisionIdentifierType,
        implicit_base: bool = False,
        inclusive: bool = False,
        assert_relative_length: bool = True,
        select_for_downgrade: bool = False,
    ) -> Iterator[Revision]:
        """Iterate through script revisions, starting at the given
        upper revision identifier and ending at the lower.

        The traversal uses strictly the `down_revision`
        marker inside each migration script, so
        it is a requirement that upper >= lower,
        else you'll get nothing back.

        The iterator yields :class:`.Revision` objects.

        """
        fn: Callable
        if select_for_downgrade:
            fn = self._collect_downgrade_revisions
        else:
            fn = self._collect_upgrade_revisions

        revisions, heads = fn(
            upper,
            lower,
            inclusive=inclusive,
            implicit_base=implicit_base,
            assert_relative_length=assert_relative_length,
        )

        for node in self._topological_sort(revisions, heads):
            yield not_none(self.get_revision(node))

    def _get_descendant_nodes(
        self,
        targets: Collection[Optional[_RevisionOrBase]],
        map_: Optional[_RevisionMapType] = None,
        check: bool = False,
        omit_immediate_dependencies: bool = False,
        include_dependencies: bool = True,
    ) -> Iterator[Any]:
        if omit_immediate_dependencies:

            def fn(rev):
                if rev not in targets:
                    return rev._all_nextrev
                else:
                    return rev.nextrev

        elif include_dependencies:

            def fn(rev):
                return rev._all_nextrev

        else:

            def fn(rev):
                return rev.nextrev

        return self._iterate_related_revisions(
            fn, targets, map_=map_, check=check
        )

    def _get_ancestor_nodes(
        self,
        targets: Collection[Optional[_RevisionOrBase]],
        map_: Optional[_RevisionMapType] = None,
        check: bool = False,
        include_dependencies: bool = True,
    ) -> Iterator[Revision]:
        if include_dependencies:

            def fn(rev):
                return rev._normalized_down_revisions

        else:

            def fn(rev):
                return rev._versioned_down_revisions

        return self._iterate_related_revisions(
            fn, targets, map_=map_, check=check
        )

    def _iterate_related_revisions(
        self,
        fn: Callable,
        targets: Collection[Optional[_RevisionOrBase]],
        map_: Optional[_RevisionMapType],
        check: bool = False,
    ) -> Iterator[Revision]:
        if map_ is None:
            map_ = self._revision_map

        seen = set()
        todo: Deque[Revision] = collections.deque()
        for target_for in targets:
            target = is_revision(target_for)
            todo.append(target)
            if check:
                per_target = set()

            while todo:
                rev = todo.pop()
                if check:
                    per_target.add(rev)

                if rev in seen:
                    continue
                seen.add(rev)
                # Check for map errors before collecting.
                for rev_id in fn(rev):
                    next_rev = map_[rev_id]
                    assert next_rev is not None
                    if next_rev.revision != rev_id:
                        raise RevisionError(
                            "Dependency resolution failed; broken map"
                        )
                    todo.append(next_rev)
                yield rev
            if check:
                overlaps = per_target.intersection(targets).difference(
                    [target]
                )
                if overlaps:
                    raise RevisionError(
                        "Requested revision %s overlaps with "
                        "other requested revisions %s"
                        % (
                            target.revision,
                            ", ".join(r.revision for r in overlaps),
                        )
                    )

    def _topological_sort(
        self,
        revisions: Collection[Revision],
        heads: Any,
    ) -> List[str]:
        """Yield revision ids of a collection of Revision objects in
        topological sorted order (i.e. revisions always come after their
        down_revisions and dependencies). Uses the order of keys in
        _revision_map to sort.

        """

        id_to_rev = self._revision_map

        def get_ancestors(rev_id):
            return {
                r.revision
                for r in self._get_ancestor_nodes([id_to_rev[rev_id]])
            }

        todo = {d.revision for d in revisions}

        # Use revision map (ordered dict) key order to pre-sort.
        inserted_order = list(self._revision_map)

        current_heads = list(
            sorted(
                {d.revision for d in heads if d.revision in todo},
                key=inserted_order.index,
            )
        )
        ancestors_by_idx = [get_ancestors(rev_id) for rev_id in current_heads]

        output = []

        current_candidate_idx = 0
        while current_heads:
            candidate = current_heads[current_candidate_idx]

            for check_head_index, ancestors in enumerate(ancestors_by_idx):
                # scan all the heads.  see if we can continue walking
                # down the current branch indicated by current_candidate_idx.
                if (
                    check_head_index != current_candidate_idx
                    and candidate in ancestors
                ):
                    current_candidate_idx = check_head_index
                    # nope, another head is dependent on us, they have
                    # to be traversed first
                    break
            else:
                # yup, we can emit
                if candidate in todo:
                    output.append(candidate)
                    todo.remove(candidate)

                # now update the heads with our ancestors.

                candidate_rev = id_to_rev[candidate]
                assert candidate_rev is not None

                heads_to_add = [
                    r
                    for r in candidate_rev._normalized_down_revisions
                    if r in todo and r not in current_heads
                ]

                if not heads_to_add:
                    # no ancestors, so remove this head from the list
                    del current_heads[current_candidate_idx]
                    del ancestors_by_idx[current_candidate_idx]
                    current_candidate_idx = max(current_candidate_idx - 1, 0)
                else:
                    if (
                        not candidate_rev._normalized_resolved_dependencies
                        and len(candidate_rev._versioned_down_revisions) == 1
                    ):
                        current_heads[current_candidate_idx] = heads_to_add[0]

                        # for plain movement down a revision line without
                        # any mergepoints, branchpoints, or deps, we
                        # can update the ancestors collection directly
                        # by popping out the candidate we just emitted
                        ancestors_by_idx[current_candidate_idx].discard(
                            candidate
                        )

                    else:
                        # otherwise recalculate it again, things get
                        # complicated otherwise.  This can possibly be
                        # improved to not run the whole ancestor thing
                        # each time but it was getting complicated
                        current_heads[current_candidate_idx] = heads_to_add[0]
                        current_heads.extend(heads_to_add[1:])
                        ancestors_by_idx[
                            current_candidate_idx
                        ] = get_ancestors(heads_to_add[0])
                        ancestors_by_idx.extend(
                            get_ancestors(head) for head in heads_to_add[1:]
                        )

        assert not todo
        return output

    def _walk(
        self,
        start: Optional[Union[str, Revision]],
        steps: int,
        branch_label: Optional[str] = None,
        no_overwalk: bool = True,
    ) -> Optional[_RevisionOrBase]:
        """
        Walk the requested number of :steps up (steps > 0) or down (steps < 0)
        the revision tree.

        :branch_label is used to select branches only when walking up.

        If the walk goes past the boundaries of the tree and :no_overwalk is
        True, None is returned, otherwise the walk terminates early.

        A RevisionError is raised if there is no unambiguous revision to
        walk to.
        """
        initial: Optional[_RevisionOrBase]
        if isinstance(start, str):
            initial = self.get_revision(start)
        else:
            initial = start

        children: Sequence[Optional[_RevisionOrBase]]
        for _ in range(abs(steps)):
            if steps > 0:
                assert initial != "base"
                # Walk up
                walk_up = [
                    is_revision(rev)
                    for rev in self.get_revisions(
                        self.bases if initial is None else initial.nextrev
                    )
                ]
                if branch_label:
                    children = self.filter_for_lineage(walk_up, branch_label)
                else:
                    children = walk_up
            else:
                # Walk down
                if initial == "base":
                    children = ()
                else:
                    children = self.get_revisions(
                        self.heads
                        if initial is None
                        else initial.down_revision
                    )
                    if not children:
                        children = ("base",)
            if not children:
                # This will return an invalid result if no_overwalk, otherwise
                # further steps will stay where we are.
                ret = None if no_overwalk else initial
                return ret
            elif len(children) > 1:
                raise RevisionError("Ambiguous walk")
            initial = children[0]

        return initial

    def _parse_downgrade_target(
        self,
        current_revisions: _RevisionIdentifierType,
        target: _RevisionIdentifierType,
        assert_relative_length: bool,
    ) -> Tuple[Optional[str], Optional[_RevisionOrBase]]:
        """
        Parse downgrade command syntax :target to retrieve the target revision
        and branch label (if any) given the :current_revisions stamp of the
        database.

        Returns a tuple (branch_label, target_revision) where branch_label
        is a string from the command specifying the branch to consider (or
        None if no branch given), and target_revision is a Revision object
        which the command refers to. target_revisions is None if the command
        refers to 'base'. The target may be specified in absolute form, or
        relative to :current_revisions.
        """
        if target is None:
            return None, None
        assert isinstance(
            target, str
        ), "Expected downgrade target in string form"
        match = _relative_destination.match(target)
        if match:
            branch_label, symbol, relative = match.groups()
            rel_int = int(relative)
            if rel_int >= 0:
                if symbol is None:
                    # Downgrading to current + n is not valid.
                    raise RevisionError(
                        "Relative revision %s didn't "
                        "produce %d migrations" % (relative, abs(rel_int))
                    )
                # Find target revision relative to given symbol.
                rev = self._walk(
                    symbol,
                    rel_int,
                    branch_label,
                    no_overwalk=assert_relative_length,
                )
                if rev is None:
                    raise RevisionError("Walked too far")
                return branch_label, rev
            else:
                relative_revision = symbol is None
                if relative_revision:
                    # Find target revision relative to current state.
                    if branch_label:
                        cr_tuple = util.to_tuple(current_revisions)
                        symbol_list: Sequence[str]
                        symbol_list = self.filter_for_lineage(
                            cr_tuple, branch_label
                        )
                        if not symbol_list:
                            # check the case where there are multiple branches
                            # but there is currently a single heads, since all
                            # other branch heads are dependent of the current
                            # single heads.
                            all_current = cast(
                                Set[Revision], self._get_all_current(cr_tuple)
                            )
                            sl_all_current = self.filter_for_lineage(
                                all_current, branch_label
                            )
                            symbol_list = [
                                r.revision if r else r  # type: ignore[misc]
                                for r in sl_all_current
                            ]

                        assert len(symbol_list) == 1
                        symbol = symbol_list[0]
                    else:
                        current_revisions = util.to_tuple(current_revisions)
                        if not current_revisions:
                            raise RevisionError(
                                "Relative revision %s didn't "
                                "produce %d migrations"
                                % (relative, abs(rel_int))
                            )
                        # Have to check uniques here for duplicate rows test.
                        if len(set(current_revisions)) > 1:
                            util.warn(
                                "downgrade -1 from multiple heads is "
                                "ambiguous; "
                                "this usage will be disallowed in a future "
                                "release."
                            )
                        symbol = current_revisions[0]
                        # Restrict iteration to just the selected branch when
                        # ambiguous branches are involved.
                        branch_label = symbol
                # Walk down the tree to find downgrade target.
                rev = self._walk(
                    start=self.get_revision(symbol)
                    if branch_label is None
                    else self.get_revision("%s@%s" % (branch_label, symbol)),
                    steps=rel_int,
                    no_overwalk=assert_relative_length,
                )
                if rev is None:
                    if relative_revision:
                        raise RevisionError(
                            "Relative revision %s didn't "
                            "produce %d migrations" % (relative, abs(rel_int))
                        )
                    else:
                        raise RevisionError("Walked too far")
                return branch_label, rev

        # No relative destination given, revision specified is absolute.
        branch_label, _, symbol = target.rpartition("@")
        if not branch_label:
            branch_label = None  # type:ignore[assignment]
        return branch_label, self.get_revision(symbol)

    def _parse_upgrade_target(
        self,
        current_revisions: _RevisionIdentifierType,
        target: _RevisionIdentifierType,
        assert_relative_length: bool,
    ) -> Tuple[Optional[_RevisionOrBase], ...]:
        """
        Parse upgrade command syntax :target to retrieve the target revision
        and given the :current_revisions stamp of the database.

        Returns a tuple of Revision objects which should be iterated/upgraded
        to. The target may be specified in absolute form, or relative to
        :current_revisions.
        """
        if isinstance(target, str):
            match = _relative_destination.match(target)
        else:
            match = None

        if not match:
            # No relative destination, target is absolute.
            return self.get_revisions(target)

        current_revisions_tup: Union[str, Tuple[Optional[str], ...], None]
        current_revisions_tup = util.to_tuple(current_revisions)

        branch_label, symbol, relative_str = match.groups()
        relative = int(relative_str)
        if relative > 0:
            if symbol is None:
                if not current_revisions_tup:
                    current_revisions_tup = (None,)
                # Try to filter to a single target (avoid ambiguous branches).
                start_revs = current_revisions_tup
                if branch_label:
                    start_revs = self.filter_for_lineage(
                        self.get_revisions(current_revisions_tup),  # type: ignore[arg-type] # noqa: E501
                        branch_label,
                    )
                    if not start_revs:
                        # The requested branch is not a head, so we need to
                        # backtrack to find a branchpoint.
                        active_on_branch = self.filter_for_lineage(
                            self._get_ancestor_nodes(
                                self.get_revisions(current_revisions_tup)
                            ),
                            branch_label,
                        )
                        # Find the tips of this set of revisions (revisions
                        # without children within the set).
                        start_revs = tuple(
                            {rev.revision for rev in active_on_branch}
                            - {
                                down
                                for rev in active_on_branch
                                for down in rev._normalized_down_revisions
                            }
                        )
                        if not start_revs:
                            # We must need to go right back to base to find
                            # a starting point for this branch.
                            start_revs = (None,)
                if len(start_revs) > 1:
                    raise RevisionError(
                        "Ambiguous upgrade from multiple current revisions"
                    )
                # Walk up from unique target revision.
                rev = self._walk(
                    start=start_revs[0],
                    steps=relative,
                    branch_label=branch_label,
                    no_overwalk=assert_relative_length,
                )
                if rev is None:
                    raise RevisionError(
                        "Relative revision %s didn't "
                        "produce %d migrations" % (relative_str, abs(relative))
                    )
                return (rev,)
            else:
                # Walk is relative to a given revision, not the current state.
                return (
                    self._walk(
                        start=self.get_revision(symbol),
                        steps=relative,
                        branch_label=branch_label,
                        no_overwalk=assert_relative_length,
                    ),
                )
        else:
            if symbol is None:
                # Upgrading to current - n is not valid.
                raise RevisionError(
                    "Relative revision %s didn't "
                    "produce %d migrations" % (relative, abs(relative))
                )
            return (
                self._walk(
                    start=self.get_revision(symbol)
                    if branch_label is None
                    else self.get_revision("%s@%s" % (branch_label, symbol)),
                    steps=relative,
                    no_overwalk=assert_relative_length,
                ),
            )

    def _collect_downgrade_revisions(
        self,
        upper: _RevisionIdentifierType,
        target: _RevisionIdentifierType,
        inclusive: bool,
        implicit_base: bool,
        assert_relative_length: bool,
    ) -> Any:
        """
        Compute the set of current revisions specified by :upper, and the
        downgrade target specified by :target. Return all dependents of target
        which are currently active.

        :inclusive=True includes the target revision in the set
        """

        branch_label, target_revision = self._parse_downgrade_target(
            current_revisions=upper,
            target=target,
            assert_relative_length=assert_relative_length,
        )
        if target_revision == "base":
            target_revision = None
        assert target_revision is None or isinstance(target_revision, Revision)

        roots: List[Revision]
        # Find candidates to drop.
        if target_revision is None:
            # Downgrading back to base: find all tree roots.
            roots = [
                rev
                for rev in self._revision_map.values()
                if rev is not None and rev.down_revision is None
            ]
        elif inclusive:
            # inclusive implies target revision should also be dropped
            roots = [target_revision]
        else:
            # Downgrading to fixed target: find all direct children.
            roots = [
                is_revision(rev)
                for rev in self.get_revisions(target_revision.nextrev)
            ]

        if branch_label and len(roots) > 1:
            # Need to filter roots.
            ancestors = {
                rev.revision
                for rev in self._get_ancestor_nodes(
                    [self._resolve_branch(branch_label)],
                    include_dependencies=False,
                )
            }
            # Intersection gives the root revisions we are trying to
            # rollback with the downgrade.
            roots = [
                is_revision(rev)
                for rev in self.get_revisions(
                    {rev.revision for rev in roots}.intersection(ancestors)
                )
            ]

            # Ensure we didn't throw everything away when filtering branches.
            if len(roots) == 0:
                raise RevisionError(
                    "Not a valid downgrade target from current heads"
                )

        heads = self.get_revisions(upper)

        # Aim is to drop :branch_revision; to do so we also need to drop its
        # descendents and anything dependent on it.
        downgrade_revisions = set(
            self._get_descendant_nodes(
                roots,
                include_dependencies=True,
                omit_immediate_dependencies=False,
            )
        )
        active_revisions = set(
            self._get_ancestor_nodes(heads, include_dependencies=True)
        )

        # Emit revisions to drop in reverse topological sorted order.
        downgrade_revisions.intersection_update(active_revisions)

        if implicit_base:
            # Wind other branches back to base.
            downgrade_revisions.update(
                active_revisions.difference(self._get_ancestor_nodes(roots))
            )

        if (
            target_revision is not None
            and not downgrade_revisions
            and target_revision not in heads
        ):
            # Empty intersection: target revs are not present.

            raise RangeNotAncestorError("Nothing to drop", upper)

        return downgrade_revisions, heads

    def _collect_upgrade_revisions(
        self,
        upper: _RevisionIdentifierType,
        lower: _RevisionIdentifierType,
        inclusive: bool,
        implicit_base: bool,
        assert_relative_length: bool,
    ) -> Tuple[Set[Revision], Tuple[Optional[_RevisionOrBase]]]:
        """
        Compute the set of required revisions specified by :upper, and the
        current set of active revisions specified by :lower. Find the
        difference between the two to compute the required upgrades.

        :inclusive=True includes the current/lower revisions in the set

        :implicit_base=False only returns revisions which are downstream
        of the current/lower revisions. Dependencies from branches with
        different bases will not be included.
        """
        targets: Collection[Revision] = [
            is_revision(rev)
            for rev in self._parse_upgrade_target(
                current_revisions=lower,
                target=upper,
                assert_relative_length=assert_relative_length,
            )
        ]

        # assert type(targets) is tuple, "targets should be a tuple"

        # Handled named bases (e.g. branch@... -> heads should only produce
        # targets on the given branch)
        if isinstance(lower, str) and "@" in lower:
            branch, _, _ = lower.partition("@")
            branch_rev = self.get_revision(branch)
            if branch_rev is not None and branch_rev.revision == branch:
                # A revision was used as a label; get its branch instead
                assert len(branch_rev.branch_labels) == 1
                branch = next(iter(branch_rev.branch_labels))
            targets = {
                need for need in targets if branch in need.branch_labels
            }

        required_node_set = set(
            self._get_ancestor_nodes(
                targets, check=True, include_dependencies=True
            )
        ).union(targets)

        current_revisions = self.get_revisions(lower)
        if not implicit_base and any(
            rev not in required_node_set
            for rev in current_revisions
            if rev is not None
        ):
            raise RangeNotAncestorError(lower, upper)
        assert (
            type(current_revisions) is tuple
        ), "current_revisions should be a tuple"

        # Special case where lower = a relative value (get_revisions can't
        # find it)
        if current_revisions and current_revisions[0] is None:
            _, rev = self._parse_downgrade_target(
                current_revisions=upper,
                target=lower,
                assert_relative_length=assert_relative_length,
            )
            assert rev
            if rev == "base":
                current_revisions = tuple()
                lower = None
            else:
                current_revisions = (rev,)
                lower = rev.revision

        current_node_set = set(
            self._get_ancestor_nodes(
                current_revisions, check=True, include_dependencies=True
            )
        ).union(current_revisions)

        needs = required_node_set.difference(current_node_set)

        # Include the lower revision (=current_revisions?) in the iteration
        if inclusive:
            needs.update(is_revision(rev) for rev in self.get_revisions(lower))
        # By default, base is implicit as we want all dependencies returned.
        # Base is also implicit if lower = base
        # implicit_base=False -> only return direct downstreams of
        # current_revisions
        if current_revisions and not implicit_base:
            lower_descendents = self._get_descendant_nodes(
                [is_revision(rev) for rev in current_revisions],
                check=True,
                include_dependencies=False,
            )
            needs.intersection_update(lower_descendents)

        return needs, tuple(targets)  # type:ignore[return-value]

    def _get_all_current(
        self, id_: Tuple[str, ...]
    ) -> Set[Optional[_RevisionOrBase]]:
        top_revs: Set[Optional[_RevisionOrBase]]
        top_revs = set(self.get_revisions(id_))
        top_revs.update(
            self._get_ancestor_nodes(list(top_revs), include_dependencies=True)
        )
        return self._filter_into_branch_heads(top_revs)


class Revision:
    """Base class for revisioned objects.

    The :class:`.Revision` class is the base of the more public-facing
    :class:`.Script` object, which represents a migration script.
    The mechanics of revision management and traversal are encapsulated
    within :class:`.Revision`, while :class:`.Script` applies this logic
    to Python files in a version directory.

    """

    nextrev: FrozenSet[str] = frozenset()
    """following revisions, based on down_revision only."""

    _all_nextrev: FrozenSet[str] = frozenset()

    revision: str = None  # type: ignore[assignment]
    """The string revision number."""

    down_revision: Optional[_RevIdType] = None
    """The ``down_revision`` identifier(s) within the migration script.

    Note that the total set of "down" revisions is
    down_revision + dependencies.

    """

    dependencies: Optional[_RevIdType] = None
    """Additional revisions which this revision is dependent on.

    From a migration standpoint, these dependencies are added to the
    down_revision to form the full iteration.  However, the separation
    of down_revision from "dependencies" is to assist in navigating
    a history that contains many branches, typically a multi-root scenario.

    """

    branch_labels: Set[str] = None  # type: ignore[assignment]
    """Optional string/tuple of symbolic names to apply to this
    revision's branch"""

    _resolved_dependencies: Tuple[str, ...]
    _normalized_resolved_dependencies: Tuple[str, ...]

    @classmethod
    def verify_rev_id(cls, revision: str) -> None:
        illegal_chars = set(revision).intersection(_revision_illegal_chars)
        if illegal_chars:
            raise RevisionError(
                "Character(s) '%s' not allowed in revision identifier '%s'"
                % (", ".join(sorted(illegal_chars)), revision)
            )

    def __init__(
        self,
        revision: str,
        down_revision: Optional[Union[str, Tuple[str, ...]]],
        dependencies: Optional[Union[str, Tuple[str, ...]]] = None,
        branch_labels: Optional[Union[str, Tuple[str, ...]]] = None,
    ) -> None:
        if down_revision and revision in util.to_tuple(down_revision):
            raise LoopDetected(revision)
        elif dependencies is not None and revision in util.to_tuple(
            dependencies
        ):
            raise DependencyLoopDetected(revision)

        self.verify_rev_id(revision)
        self.revision = revision
        self.down_revision = tuple_rev_as_scalar(util.to_tuple(down_revision))
        self.dependencies = tuple_rev_as_scalar(util.to_tuple(dependencies))
        self._orig_branch_labels = util.to_tuple(branch_labels, default=())
        self.branch_labels = set(self._orig_branch_labels)

    def __repr__(self) -> str:
        args = [repr(self.revision), repr(self.down_revision)]
        if self.dependencies:
            args.append("dependencies=%r" % (self.dependencies,))
        if self.branch_labels:
            args.append("branch_labels=%r" % (self.branch_labels,))
        return "%s(%s)" % (self.__class__.__name__, ", ".join(args))

    def add_nextrev(self, revision: Revision) -> None:
        self._all_nextrev = self._all_nextrev.union([revision.revision])
        if self.revision in revision._versioned_down_revisions:
            self.nextrev = self.nextrev.union([revision.revision])

    @property
    def _all_down_revisions(self) -> Tuple[str, ...]:
        return util.dedupe_tuple(
            util.to_tuple(self.down_revision, default=())
            + self._resolved_dependencies
        )

    @property
    def _normalized_down_revisions(self) -> Tuple[str, ...]:
        """return immediate down revisions for a rev, omitting dependencies
        that are still dependencies of ancestors.

        """
        return util.dedupe_tuple(
            util.to_tuple(self.down_revision, default=())
            + self._normalized_resolved_dependencies
        )

    @property
    def _versioned_down_revisions(self) -> Tuple[str, ...]:
        return util.to_tuple(self.down_revision, default=())

    @property
    def is_head(self) -> bool:
        """Return True if this :class:`.Revision` is a 'head' revision.

        This is determined based on whether any other :class:`.Script`
        within the :class:`.ScriptDirectory` refers to this
        :class:`.Script`.   Multiple heads can be present.

        """
        return not bool(self.nextrev)

    @property
    def _is_real_head(self) -> bool:
        return not bool(self._all_nextrev)

    @property
    def is_base(self) -> bool:
        """Return True if this :class:`.Revision` is a 'base' revision."""

        return self.down_revision is None

    @property
    def _is_real_base(self) -> bool:
        """Return True if this :class:`.Revision` is a "real" base revision,
        e.g. that it has no dependencies either."""

        # we use self.dependencies here because this is called up
        # in initialization where _real_dependencies isn't set up
        # yet
        return self.down_revision is None and self.dependencies is None

    @property
    def is_branch_point(self) -> bool:
        """Return True if this :class:`.Script` is a branch point.

        A branchpoint is defined as a :class:`.Script` which is referred
        to by more than one succeeding :class:`.Script`, that is more
        than one :class:`.Script` has a `down_revision` identifier pointing
        here.

        """
        return len(self.nextrev) > 1

    @property
    def _is_real_branch_point(self) -> bool:
        """Return True if this :class:`.Script` is a 'real' branch point,
        taking into account dependencies as well.

        """
        return len(self._all_nextrev) > 1

    @property
    def is_merge_point(self) -> bool:
        """Return True if this :class:`.Script` is a merge point."""

        return len(self._versioned_down_revisions) > 1


@overload
def tuple_rev_as_scalar(rev: None) -> None:
    ...


@overload
def tuple_rev_as_scalar(
    rev: Union[Tuple[_T, ...], List[_T]]
) -> Union[_T, Tuple[_T, ...], List[_T]]:
    ...


def tuple_rev_as_scalar(
    rev: Optional[Sequence[_T]],
) -> Union[_T, Sequence[_T], None]:
    if not rev:
        return None
    elif len(rev) == 1:
        return rev[0]
    else:
        return rev


def is_revision(rev: Any) -> Revision:
    assert isinstance(rev, Revision)
    return rev



# mod_97_10.py - functions for performing the ISO 7064 Mod 97, 10 algorithm
#
# Copyright (C) 2010-2022 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""The ISO 7064 Mod 97, 10 algorithm.

The Mod 97, 10 algorithm evaluates the whole number as an integer which is
valid if the number modulo 97 is 1. As such it has two check digits.

>>> calc_check_digits('99991234567890121414')
'90'
>>> validate('9999123456789012141490')
'9999123456789012141490'
>>> calc_check_digits('4354111611551114')
'31'
>>> validate('08686001256515001121751')
'08686001256515001121751'
>>> calc_check_digits('22181321402534321446701611')
'35'
"""

from stdnum.exceptions import *


def _to_base10(number):
    """Prepare the number to its base10 representation."""
    return ''.join(
        str(int(x, 36)) for x in number)


def checksum(number):
    """Calculate the checksum. A valid number should have a checksum of 1."""
    return int(_to_base10(number)) % 97


def calc_check_digits(number):
    """Calculate the extra digits that should be appended to the number to
    make it a valid number."""
    return '%02d' % (98 - checksum(number + '00'))


def validate(number):
    """Check whether the check digit is valid."""
    try:
        valid = checksum(number) == 1
    except Exception:  # noqa: B902
        raise InvalidFormat()
    if not valid:
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check whether the check digit is valid."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



class ErrorBuildingDataBuilder(Exception):

    def __init__(
        self,
        report_name: str,
        error_message: str,
    ) -> None:
        message = f"Error building data builder for report {report_name}: {error_message}"
        super(ErrorBuildingDataBuilder, self).__init__(message)
